from parsy import regex, seq, string

space = regex(r"\s+")  # non-optional whitespace
padding = regex(r"\s*")  # optional whitespace
anything = regex(".*")
identifier = regex("[a-zA-Z][a-zA-Z0-9_]*")
tableidentifier = seq(keyspace=identifier << string("."), table=identifier) | seq(
    table=identifier
)


def lexeme(p):
    return p << padding


def ci_str_token(input_str):
    return lexeme(string(input_str, transform=lambda s: s.upper()))



import logging
from typing import List, Optional, Tuple

from ..datatypes.common import FlowDirection
from ..ingest.utxo import (
    BtcStreamerAdapter,
    OutputResolverBase,
    enrich_txs,
    get_stream_adapter,
)
from ..utils import parse_timestamp
from ..utils.logging import suppress_log_level
from .abstract import FlowEvent, FlowProvider

logger = logging.getLogger(__name__)


def parse_btcetl_txs(tx) -> List[FlowEvent]:
    output = []
    tx_h = tx["hash"]
    time = parse_timestamp(tx["block_timestamp"])
    b = tx["block_number"]

    for o in tx["outputs"]:
        v = (
            (o["value"] / len(o["addresses"]))
            if len(o["addresses"]) > 0
            else o["value"]
        )
        for a in o.get("addresses", []):
            output.append(
                FlowEvent(
                    direction=FlowDirection.OUT,
                    address=a,
                    value=v,
                    block=b,
                    tx_ref=tx_h,
                    timestamp=time,
                )
            )

    for o in tx["inputs"]:
        if "addresses" in o and len(o["addresses"]) != 0:
            v = (
                (o["value"] / len(o["addresses"]))
                if len(o["addresses"]) > 0
                else o["value"]
            )
        for a in o.get("addresses", []):
            output.append(
                FlowEvent(
                    direction=FlowDirection.OUT,
                    address=a,
                    value=v,
                    block=b,
                    tx_ref=tx_h,
                    timestamp=time,
                )
            )
    return output


class BitcoinEtlFlowProvider(FlowProvider):
    node_url: str
    stream: BtcStreamerAdapter
    output_resolver: OutputResolverBase

    def __init__(
        self, currency: str, node_url: str, output_resolver: OutputResolverBase
    ):
        self.node_url = node_url
        self.stream = get_stream_adapter(currency, node_url, batch_size=1)
        self.output_resolver = output_resolver

    def __enter__(self):
        self.output_resolver.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.output_resolver.__exit__(exc_type, exc_val, exc_tb)

    def get_flows_for_block(
        self, block: int
    ) -> Optional[List[Tuple[FlowEvent, object]]]:
        with suppress_log_level(logging.INFO):
            txs = self.stream.export_transactions(block, block)
            enrich_txs(txs, self.output_resolver, ignore_missing_outputs=True)

        events = []
        for tx in txs:
            for e in parse_btcetl_txs(tx):
                events.append((e, tx))

        return events if events else None



"""This file contains all db functions that are specific to the graphsense database.
Functions that have specific implementations based on the currency are stored
in the eth.py and btc.py files. Generic database functions belong in cassandra.py

Attributes:
    DATE_FORMAT (str): Format string of date format used by the database (str)
"""

import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from functools import lru_cache, partial
from typing import Iterable, List, Optional, Sequence, Tuple, Union

import pandas as pd
from cassandra import WriteTimeout
from tenacity import Retrying, retry_if_exception_type, stop_after_attempt

from ..config import keyspace_types
from ..datatypes import DbChangeType
from ..utils import GenericArrayFacade, binary_search, parse_timestamp
from ..utils.account import get_id_group, get_id_group_with_secondary_relations
from .cassandra import (
    CassandraDb,
    build_create_stmt,
    build_delete_stmt,
    build_insert_stmt,
    build_select_stmt,
    build_truncate_stmt,
)

CONCURRENCY = 2000
DATE_FORMAT = "%Y-%m-%d"
FIRST_BLOCK = defaultdict(lambda: 0)
FIRST_BLOCK["trx"] = 1
logger = logging.getLogger(__name__)


@dataclass
class DbChange:
    action: DbChangeType
    table: str
    data: dict

    @classmethod
    def new(Cls, table: str, data: dict):
        return Cls(action=DbChangeType.NEW, table=table, data=data)

    @classmethod
    def update(Cls, table: str, data: dict):
        return Cls(action=DbChangeType.UPDATE, table=table, data=data)

    @classmethod
    def truncate(Cls, table: str):
        return Cls(action=DbChangeType.TRUNCATE, table=table, data=None)

    @classmethod
    def delete(Cls, table: str, data: dict):
        return Cls(action=DbChangeType.DELETE, table=table, data=data)

    def get_cql_statement(self, keyspace):
        if self.action == DbChangeType.UPDATE or self.action == DbChangeType.NEW:
            return build_insert_stmt(
                columns=self.data.keys(), table=self.table, keyspace=keyspace
            )
        elif self.action == DbChangeType.TRUNCATE:
            return build_truncate_stmt(table=self.table, keyspace=keyspace)
        elif self.action == DbChangeType.DELETE:
            return build_delete_stmt(
                table=self.table, key_columns=self.data.keys(), keyspace=keyspace
            )
        else:
            raise Exception(
                f"Don't know how to build statement for action {self.action}."
            )


class KeyspaceConfig:
    def __init__(self, keyspace_name, db_type, address_type, tx_hash_type, currency):
        self._keyspace_name = keyspace_name
        self._db_type = db_type
        self._address_type = address_type
        self._tx_hash_type = tx_hash_type
        self._currency = currency

    @property
    def keyspace_name(self):
        return self._keyspace_name

    @property
    def db_type(self):
        return self._db_type

    @property
    def address_type(self):
        return self._address_type

    @property
    def tx_hash_type(self):
        return self._tx_hash_type


class WithinKeyspace:
    def select_stmt(
        self,
        table: str,
        columns: Sequence[str] = ["*"],
        where: Optional[dict] = None,
        limit: Optional[int] = None,
        per_partition_limit: Optional[int] = None,
    ) -> str:
        return build_select_stmt(
            table,
            columns=columns,
            keyspace=self.get_keyspace(),
            where=where,
            limit=limit,
            per_partition_limit=per_partition_limit,
        )

    def get_keyspace(self):
        return self._keyspace


class DbReaderMixin:
    def select_safe(
        self,
        table: str,
        columns: Sequence[str] = ["*"],
        where: Optional[dict] = None,
        limit: Optional[int] = None,
        per_partition_limit: Optional[int] = None,
        fetch_size=None,
    ):
        return self._db.execute_safe(
            self.select_stmt(
                table=table,
                columns=columns,
                where={k: f"%({k})s" for k, v in where.items()},
                limit=limit,
                per_partition_limit=per_partition_limit,
            ),
            where,
            fetch_size=fetch_size,
        )

    def select(
        self,
        table: str,
        columns: Sequence[str] = ["*"],
        where: Optional[dict] = None,
        limit: Optional[int] = None,
        per_partition_limit: Optional[int] = None,
        fetch_size=None,
    ):
        return self._db.execute(
            self.select_stmt(
                table=table,
                columns=columns,
                where=where,
                limit=limit,
                per_partition_limit=per_partition_limit,
            ),
            fetch_size=fetch_size,
        )

    def select_one(
        self, table: str, columns: Sequence[str] = ["*"], where: Optional[dict] = None
    ):
        return self._at_most_one_result(
            self.select(table=table, columns=columns, where=where, limit=2)
        )

    def select_one_safe(
        self, table: str, columns: Sequence[str] = ["*"], where: Optional[dict] = None
    ):
        return self._at_most_one_result(
            self.select_safe(table=table, columns=columns, where=where, limit=2)
        )

    def get_columns_for_table(self, table: str):
        return self._db.get_columns_for_table(self.get_keyspace(), table)

    def select_async(
        self,
        table: str,
        columns: Sequence[str] = ["*"],
        where: Optional[dict] = None,
        limit: Optional[int] = None,
        per_partition_limit: Optional[int] = None,
        fetch_size=None,
    ):
        return self._db.execute_async(
            self.select_stmt(
                table=table,
                columns=columns,
                where=where,
                limit=limit,
                per_partition_limit=per_partition_limit,
            ),
            fetch_size=fetch_size,
        )

    def select_async_safe(
        self,
        table: str,
        columns: Sequence[str] = ["*"],
        where: Optional[dict] = None,
        limit: Optional[int] = None,
        per_partition_limit: Optional[int] = None,
        fetch_size=None,
    ):
        return self._db.execute_async_safe(
            self.select_stmt(
                table=table,
                columns=columns,
                where={k: f"%({k})s" for k, v in where.items()},
                limit=limit,
                per_partition_limit=per_partition_limit,
            ),
            where,
            fetch_size=fetch_size,
        )

    def _get_hightest_id(
        self, table="block", sanity_check=True, id_col=None
    ) -> Optional[int]:
        """Return last ingested address ID from a table."""

        if id_col is None:
            id_col = f"{table}_id"

        group_id_col = f"{id_col}_group"

        result = self.select(table=table, columns=[group_id_col], per_partition_limit=1)
        groups = [getattr(row, group_id_col) for row in result.current_rows]

        if len(groups) > 0:
            result = self.select(
                table,
                columns=[f"MAX({id_col}) AS max"],
                where={group_id_col: max(groups)},
            )
            highest_id = int(result.current_rows[0].max)

            if sanity_check:
                self._ensure_is_highest_id(table, id_col, highest_id)

            return highest_id
        else:
            return None

    def _get_bucket_divisors_by_table_name(self) -> dict:
        raise Exception("Must be implemented in chain specific child class")

    def _ensure_is_highest_id(
        self, table: str, id_col: str, query_id: Optional[int]
    ) -> bool:
        if query_id is None:
            return None

        groups = self._get_bucket_divisors_by_table_name()

        if table in groups:
            # this case handles tables with group ids.
            group_id_col = f"{id_col}_group"
            bucket_divisor = groups[table]
            highest_plus_one = query_id + 1

            id_group_highest_plus_one = self.get_id_group(
                highest_plus_one, bucket_divisor
            )

            w = {group_id_col: id_group_highest_plus_one, id_col: highest_plus_one}
        else:
            # this case handles tables with no group column and increasing integer ids.
            w = {id_col: query_id + 1}
        result = self.select(table, columns=[id_col], where=w)
        if len(result.current_rows) > 0:
            raise Exception(
                (
                    f"Something went wrong {query_id} "
                    " is not the highest_id"
                    f" {query_id+1} exist in {table}.{id_col}"
                )
            )

    def _at_most_one_result(self, result):
        if len(result.current_rows) > 1:
            raise Exception("Config tables are supposed to only have exactly one row.")
        return result.current_rows[0] if len(result.current_rows) > 0 else None

    def _get_only_row_from_table(self, table: str = "configuration"):
        return self._at_most_one_result(self.select(table, limit=2))

    def get_id_group(self, id_, bucket_size):
        return get_id_group(id_, bucket_size)


class DbWriterMixin:
    """
    This mixin requires the object to provide a CassandraDB instance at
    self._db and requires the object to provide the WithinKeyspace mixin
    """

    def apply_changes(self, changes: List[DbChange], atomic=True, nr_retries=5):
        statements = [
            (chng.get_cql_statement(keyspace=self.get_keyspace()), chng)
            for chng in changes
        ]
        unique_statements = {stmt for stmt, _ in statements}

        prepared_statements = {
            stmt: self._db.get_prepared_statement(stmt) for stmt in unique_statements
        }

        change_stmts = [
            prepared_statements[
                chng.get_cql_statement(keyspace=self.get_keyspace())
            ].bind(chng.data)
            for chng in changes
        ]

        attempts_made = 0
        for attempt in Retrying(
            retry=retry_if_exception_type(WriteTimeout),
            reraise=True,
            stop=stop_after_attempt(nr_retries),
        ):
            # see https://tenacity.readthedocs.io/en/latest/#retrying-code-block
            with attempt:
                attempts_made += 1
                if attempts_made > 1:
                    logger.warning(
                        "Applying changes ran into a write timeout. "
                        f"Retrying {(nr_retries  - attempts_made) +1} more times."
                    )
                if atomic:
                    self._db.execute_statements_atomic(change_stmts)
                else:
                    self._db.execute_statements(change_stmts)

    def ensure_table_exists(
        self,
        table_name: str,
        columns: Sequence[str],
        pk_columns: Sequence[str],
        truncate: bool = False,
    ):
        self._db.execute(
            build_create_stmt(
                columns,
                pk_columns,
                table_name,
                fail_if_exists=False,
                keyspace=self.get_keyspace(),
            )
        )
        if truncate:
            self._db.execute(
                build_truncate_stmt(table_name, keyspace=self.get_keyspace())
            )

    def ingest(
        self,
        table: str,
        items: Iterable,
        upsert=True,
        cl=None,
        concurrency: int = 100,
        auto_none_to_unset=False,
    ):
        self._db.ingest(
            table,
            self.get_keyspace(),
            items,
            upsert=upsert,
            cl=cl,
            concurrency=concurrency,
            auto_none_to_unset=auto_none_to_unset,
        )


def get_last_notnone(result, start, end):
    for i in range(end, start, -1):
        if result[i] is not None:
            return i
    return -1


class RawDb(ABC, WithinKeyspace, DbReaderMixin, DbWriterMixin):
    def __init__(self, keyspace_config: KeyspaceConfig, db: CassandraDb):
        self._keyspace_config = keyspace_config
        self._keyspace = keyspace_config.keyspace_name
        self._db = db

    def _get_bucket_divisors_by_table_name(self) -> dict:
        return {
            "block": self.get_block_bucket_size(),
            "transaction": self.get_tx_bucket_size(),
        }

    def exists(self) -> bool:
        return self._db.has_keyspace(self._keyspace)

    def keyspace_name(self) -> str:
        return self._keyspace

    def get_summary_statistics(self) -> Optional[object]:
        if not self.exists():
            return None
        if self._db.has_table(self._keyspace, "summary_statistics"):
            return self._get_only_row_from_table("summary_statistics")
        else:
            return None

    def find_block_nr_for_date(self, date: datetime) -> int:
        hb = self.get_highest_block()
        start = 0

        def get_item(date, index):
            daq = self.get_block_timestamp(index)
            return 0 if daq <= date else 1

        get_item_date = partial(get_item, date)

        r = binary_search(GenericArrayFacade(get_item_date), 1, lo=start, hi=hb)
        # r = get_last_notnone(GenericArrayFacade(get_item_date), start, hb)

        if r == -1:
            # minus one could mean two things, either
            # no missing exchangesrates are found within the lookback range
            # or that all have exchange rates, so we recheck if the
            # second case applies
            er = self.get_block_timestamp(hb)
            if er is not None:
                r = hb
        else:
            r = r - 1

        return r

    def find_highest_block_with_exchange_rates(
        self, lookback_in_blocks=86400, validate=True
    ) -> int:
        """Summary
            Searches for the highest block with available exchange_rates in
            the raw keyspace this is the maximum block the importer works on.

        Args:
            lookback_in_blocks (int, optional): how many blocks to look in the past.
                                                Default 86400 so approx.
                                                13 days on ethereum pace
        Returns:
            int: Last block that exchangerates are available
        """
        hb = self.get_highest_block()

        start = max(hb - lookback_in_blocks, 0)

        def has_er_value(result, i=0):
            return result[i]["fiat_values"] is not None

        def get_item(index):
            batch = self.get_exchange_rates_for_block_batch([index])
            return 0 if has_er_value(batch) else 1

        def get_item_first_rates(index):
            batch = self.get_exchange_rates_for_block_batch([index])
            return 1 if has_er_value(batch) else 0

        # find first block with exchange rates
        first_rates = max(
            binary_search(GenericArrayFacade(get_item_first_rates), 1, lo=start, hi=hb),
            0,
        )
        r = binary_search(GenericArrayFacade(get_item), 1, lo=first_rates, hi=hb)

        if r == -1:
            # minus one could mean two things, either
            # no missing exchangesrates are found within the lookback range
            # or that all have exchange rates, so we recheck if the
            # second case applies
            er = self.get_exchange_rates_for_block_batch([hb])
            if len(er) > 0 and has_er_value(er):
                r = hb
        else:
            r = r - 1

        if validate and r != -1:
            # ers = self.get_exchange_rates_for_block_batch([r - 1, r, r + 1])
            ers = self.get_exchange_rates_for_block_batch([r - 1, r])
            assert has_er_value(ers, i=0) and has_er_value(ers, i=1)
            # if has_er_value(ers, i=2):
            #     logger.warning(f"Found exchange-rate for "
            # "not yet imported block {r+1}")

        return r

    @lru_cache(maxsize=1)
    def get_configuration(self) -> Optional[object]:
        return self._get_only_row_from_table("configuration")

    def is_configuration_populated(self) -> bool:
        return self._get_only_row_from_table("configuration") is not None

    def get_exchange_rates(self, table=None) -> Iterable:
        r = self.select("exchange_rates" if table is None else table)
        return list(r)

    def get_last_exchange_rate_date(self, table=None) -> Optional[datetime]:
        ra = self.get_exchange_rates(table=table)
        return (
            max([datetime.fromisoformat(r.date) for r in ra]) if len(ra) > 0 else None
        )

    def get_exchange_rates_by_date(self, date) -> Iterable:
        date_str = date.strftime(DATE_FORMAT)
        r = self.select("exchange_rates", where={"date": date_str})
        return self._at_most_one_result(r)

    def get_block_bucket_size(self) -> Optional[int]:
        config = self.get_configuration()
        return int(config.block_bucket_size) if config is not None else None

    def get_tx_bucket_size(self) -> Optional[int]:
        config = self.get_configuration()
        return (
            int(config.tx_bucket_size)
            if config is not None and "tx_bucket_size" in dir(config)
            else None
        )

    def get_tx_prefix_length(self) -> Optional[int]:
        config = self.get_configuration()
        return (
            int(config.tx_prefix_length)
            if config is not None and "tx_prefix_length" in dir(config)
            else None
        )

    def get_highest_block(self, sanity_check=True) -> Optional[int]:
        """Return last ingested block ID from block table."""
        if self.exists():
            return self._get_hightest_id(table="block", sanity_check=sanity_check)
        else:
            return None

    @abstractmethod
    def get_transaction_ids_in_block(self, block: int) -> Iterable:
        raise Exception("Must be implemented in chain specific child class")

    @abstractmethod
    def get_addresses_in_block(self, block: int) -> Iterable:
        raise Exception("Must be implemented in chain specific child class")

    @abstractmethod
    def get_transactions_in_block(self, block: int) -> Iterable:
        raise Exception("Must be implemented in chain specific child class")

    def get_block_timestamps_batch(self, blocks: list[int]):
        bucket_size = self.get_block_bucket_size()
        stmt = self.select_stmt(
            "block",
            columns=["timestamp"],
            where={"block_id_group": "?", "block_id": "?"},
            limit=1,
        )
        parameters = [(b, [self.get_id_group(b, bucket_size), b]) for b in blocks]
        results = self._db.execute_batch(stmt, parameters)
        return {
            a: (
                (parse_timestamp(row.current_rows[0].timestamp))
                if len(row.current_rows) > 0
                else None
            )
            for (a, row) in results
        }

    def get_block_timestamp(self, block: int) -> datetime:
        btc = self.get_block_timestamps_batch([block])
        return btc[block]

    def get_exchange_rates_batch(self, dates: list[datetime]):
        stmt = self.select_stmt("exchange_rates", where={"date": "?"}, limit=1)
        ds = list({date.strftime(DATE_FORMAT) for date in dates if date is not None})
        results = self._db.execute_batch(stmt, [(d, [d]) for d in ds])

        # TODO maybe refactor with _at_most_one?
        def get_result_item(r):
            r = r.current_rows
            return r[0] if len(r) > 0 else None

        return {a: get_result_item(row) for (a, row) in results}

    def get_exchange_rates_for_block_batch(self, batch: list[int]):
        block_to_ts = self.get_block_timestamps_batch(batch)
        exchange_rate_to_date = self.get_exchange_rates_batch(block_to_ts.values())

        def get_values_list(er):
            return (
                [er.fiat_values.get(x) for x in list(er.fiat_values.keys())]
                if er is not None
                else None
            )

        ers = [
            (
                b,
                (
                    exchange_rate_to_date[block_to_ts[b].strftime(DATE_FORMAT)]
                    if block_to_ts[b] is not None
                    else None
                ),
            )
            for b in batch
        ]
        return [{"block_id": b, "fiat_values": get_values_list(er)} for b, er in ers]


class TransformedDb(ABC, WithinKeyspace, DbReaderMixin, DbWriterMixin):
    def __init__(self, keyspace_config: KeyspaceConfig, db: CassandraDb):
        self._keyspace_config = keyspace_config
        self._keyspace = keyspace_config.keyspace_name
        self._db = db
        self._db_config = None

    def _get_bucket_divisors_by_table_name(self) -> dict:
        bucket_size = self.get_address_id_bucket_size()
        return {
            "address": bucket_size,
            "cluster": self.get_cluster_id_bucket_size(),
            "transaction_ids_by_transaction_id_group": bucket_size,
        }

    def exists(self) -> bool:
        return self._db.has_keyspace(self._keyspace)

    def keyspace_name(self) -> str:
        return self._keyspace

    def get_summary_statistics(self) -> Optional[object]:
        if not self.exists():
            return None
        return self._get_only_row_from_table("summary_statistics")

    def get_exchange_rates_by_block(self, block) -> Iterable:
        return self.select_one("exchange_rates", where={"block_id": block})

    def get_address_id_bucket_size(self) -> Optional[int]:
        config = self.get_configuration()
        return int(config.bucket_size) if config is not None else None

    def get_block_id_bucket_size(self) -> Optional[int]:
        config = self.get_configuration()
        return int(config.bucket_size) if config is not None else None

    def get_address_transactions_id_bucket_size(self) -> Optional[int]:
        config = self.get_configuration()
        return int(config.block_bucket_size_address_txs) if config is not None else None

    def get_addressrelations_ids_nbuckets(self) -> Optional[int]:
        config = self.get_configuration()
        return int(config.addressrelations_ids_nbuckets) if config is not None else None

    def get_cluster_id_bucket_size(self) -> Optional[int]:
        return self.get_address_id_bucket_size()

    def get_configuration(self) -> Optional[object]:
        if self._db_config is None:
            self._db_config = self._get_only_row_from_table("configuration")
        return self._db_config

    def is_configuration_populated(self) -> bool:
        return self._get_only_row_from_table("configuration") is not None

    def get_highest_address_id(self, sanity_check=True) -> Optional[int]:
        """Return last ingested address ID from address table."""
        du = self.get_last_delta_updater_state()
        ha = self._get_hightest_id(table="address", sanity_check=sanity_check)
        return max(ha or 0, du.highest_address_id) if du is not None else ha

    def get_highest_transaction_id(self):
        return None

    @abstractmethod
    def get_highest_cluster_id(self, sanity_check=True) -> Optional[int]:
        raise Exception("Must be implemented in chain specific child class")

    def get_highest_block(self) -> Optional[int]:
        stats = self.get_summary_statistics()
        if stats is None:
            return None

        # minus one when starting to count at 0
        height_minus_noblocks = FIRST_BLOCK[self._keyspace_config._currency] - 1
        height = height_minus_noblocks + int(stats.no_blocks)

        return height

    def is_first_delta_update_run(self) -> bool:
        stats = self.get_summary_statistics()
        return stats is not None and int(stats.no_blocks) == int(
            stats.no_blocks_transform
        )

    def get_highest_block_fulltransform(self) -> Optional[int]:
        stats = self.get_summary_statistics()
        height_minus_noblocks = FIRST_BLOCK[self._keyspace_config._currency] - 1
        if stats is None:
            return None

        if hasattr(stats, "no_blocks_transform"):
            return height_minus_noblocks + int(stats.no_blocks_transform)
        else:
            return height_minus_noblocks + int(stats.no_blocks)

    def get_highest_exchange_rate_block(self, sanity_check=True) -> Optional[int]:
        res = self.select("exchange_rates", columns=["block_id"], per_partition_limit=1)
        m = max([x.block_id for x in res])
        if sanity_check:
            self._ensure_is_highest_id("exchange_rates", "block_id", m)
        return m

    def get_highest_block_delta_updater(self, sanity_check=True) -> Optional[int]:
        hb = self.get_highest_block()
        du = self.get_last_delta_updater_state()
        if du is not None and sanity_check:
            self._ensure_is_highest_id(
                "delta_updater_history", "last_synced_block", du.last_synced_block
            )

        return int(du.last_synced_block) if du is not None else hb

    def get_delta_updater_history(self) -> Iterable:
        if self._db.has_table(self._keyspace, "delta_updater_history"):
            return list(self.select("delta_updater_history", fetch_size=100))
        else:
            return []

    def get_last_delta_updater_state(self):
        if self._db.has_table(self._keyspace, "delta_updater_status"):
            return self._get_only_row_from_table("delta_updater_status")
        else:
            return None

    def get_address_prefix_length(self) -> Optional[int]:
        config = self.get_configuration()
        return int(config.address_prefix_length) if config is not None else None

    def to_db_address(self, address):
        Address = self._keyspace_config.address_type
        return Address(address, self.get_configuration())

    def to_db_tx_hash(self, tx_hash):
        TxHash = self._keyspace_config.tx_hash_type
        return TxHash(tx_hash, self.get_configuration())

    @lru_cache(maxsize=1_000_000)
    def knows_address(self, address: Union[str, bytearray]) -> bool:
        """Checks if address is in transformed keyspace.
        uses lru cache of 1_000_000 items. This should speedup lookups.
        Conservative estimate is that a full cache is about 130 MB

        Args:
            address (Union[str, bytearray]): Address as either bytearray or
                                             hex string.

        Returns:
            bool: True if present
        """
        adr = self.to_db_address(address)
        w = {"address_prefix": f"'{adr.prefix}'", "address": adr.db_encoding_query}
        rows = self.select(
            "address_ids_by_address_prefix", columns=["address"], where=w, limit=1
        )
        return len(rows.current_rows) > 0

    def known_addresses_batch(
        self, address: list[str], inlcude_new_adresses_table=True
    ):
        stmt = self.select_stmt(
            "address_ids_by_address_prefix",
            columns=["address"],
            where={"address_prefix": "?", "address": "?"},
            limit=1,
        )
        adrs = [(a, self.to_db_address(a)) for a in address]
        parameters = [(a, [adr.prefix, adr.db_encoding]) for a, adr in adrs]

        results = self._db.execute_batch_async(stmt, parameters)
        if inlcude_new_adresses_table:
            stmt = self.select_stmt(
                "new_addresses",
                columns=["address"],
                where={"address_prefix": "?", "address": "?"},
                limit=1,
            )
            results_new = self._db.execute_batch_async(stmt, parameters)
            r1 = {
                a: (len(row.current_rows) > 0)
                for (a, row) in self._db.await_batch(results)
            }
            r2 = {
                a: (len(row.current_rows) > 0)
                for (a, row) in self._db.await_batch(results_new)
            }

            return {a: (r or r2[a]) for (a, r) in r1.items()}

        else:
            return {
                a: (len(row.current_rows) > 0)
                for (a, row) in self._db.await_batch(results)
            }

    def get_token_configuration(self):
        stmt = self.select_stmt("token_configuration", limit=100)
        res = self._db.execute(stmt)
        df = pd.DataFrame(res)
        df["token_address"] = df["token_address"].apply(lambda x: "0x" + x.hex())
        return df

    def get_address_id_async_batch(self, addresses: List[str]):
        stmt = self.select_stmt(
            "address_ids_by_address_prefix",
            columns=["*"],
            where={k: "?" for k in ["address_prefix", "address"]},
            limit=1,
        )
        prep = self._db.get_prepared_statement(stmt)

        bstmts = [
            prep.bind({"address_prefix": f"{adr.prefix}", "address": adr.db_encoding})
            for adr in [self.to_db_address(address) for address in addresses]
        ]

        return zip(
            addresses,
            self._db.execute_statements_async(bstmts, concurrency=CONCURRENCY),
        )

    def get_address_id_async(self, address: str):
        adr = self.to_db_address(address)
        w = {"address_prefix": f"'{adr.prefix}'", "address": adr.db_encoding_query}
        return self.select_async(
            "address_ids_by_address_prefix",
            columns=["*"],
            where=w,
            limit=1,
        )

    def get_address_async_batch(self, address_ids: List[int]):
        bs = self.get_address_id_bucket_size()
        stmt = self.select_stmt(
            "address",
            columns=["*"],
            where={k: "?" for k in ["address_id_group", "address_id"]},
            limit=1,
        )
        prep = self._db.get_prepared_statement(stmt)

        bstmts = [
            prep.bind(
                {
                    "address_id_group": self.get_id_group(addr_id, bs),
                    "address_id": addr_id,
                }
            )
            for addr_id in address_ids
        ]

        return zip(
            address_ids,
            self._db.execute_statements_async(bstmts, concurrency=CONCURRENCY),
        )

    def get_address_async(self, address_id: int):
        bucket = self.get_id_group(address_id, self.get_address_id_bucket_size())
        w = {"address_id_group": bucket, "address_id": f"{address_id}"}
        return self.select_async(
            "address",
            columns=["*"],
            where=w,
            limit=1,
        )

    def get_address_incoming_relations_async_batch(
        self, rel_ids: List[Tuple[int, int]]
    ):
        stmt = self.select_stmt(
            "address_incoming_relations",
            columns=["*"],
            where={
                k: "?"
                for k in ["dst_address_id_group", "dst_address_id", "src_address_id"]
            },
            limit=1,
        )
        prep = self._db.get_prepared_statement(stmt)

        bstmts = [
            prep.bind(
                {
                    "dst_address_id_group": self.get_id_group(
                        dst_address, self.get_address_id_bucket_size()
                    ),
                    "dst_address_id": dst_address,
                    "src_address_id": src_address,
                }
            )
            for dst_address, src_address in rel_ids
        ]

        return self._db.execute_statements_async(bstmts, concurrency=CONCURRENCY)

    def get_address_inrelations_async_batch_account(
        self, rel_ids: List[Tuple[int, int]]
    ):
        stmt = self.select_stmt(
            "address_incoming_relations",
            columns=["*"],
            where={
                k: "?"
                for k in [
                    "dst_address_id_group",
                    "dst_address_id_secondary_group",
                    "dst_address_id",
                    "src_address_id",
                ]
            },
            limit=1,
        )
        prep = self._db.get_prepared_statement(stmt)

        bucketsize = self.get_address_id_bucket_size()
        relations_nbuckets = self.get_addressrelations_ids_nbuckets()

        bstmts = []
        for dst_address, src_address in rel_ids:
            address_group, secondary_group = get_id_group_with_secondary_relations(
                dst_address, src_address, bucketsize, relations_nbuckets
            )
            bstmts.append(
                prep.bind(
                    {
                        "dst_address_id_group": address_group,
                        "dst_address_id_secondary_group": secondary_group,
                        "dst_address_id": dst_address,
                        "src_address_id": src_address,
                    }
                )
            )

        return self._db.execute_statements_async(bstmts, concurrency=CONCURRENCY)

    def get_address_incoming_relations_async(
        self, address_id: int, src_address_id: Optional[int]
    ):
        w = {
            "dst_address_id_group": self.get_id_group(
                address_id, self.get_address_id_bucket_size()
            ),
            "dst_address_id": address_id,
        }
        if src_address_id is not None:
            w["src_address_id"] = src_address_id
        return self.select_async(
            "address_incoming_relations",
            columns=["*"],
            where=w,
            limit=1,
        )

    def get_max_secondary_ids_async(
        self, address_id_groups: List[int], tablename: str, id_group_col: str
    ):
        stmt = self.select_stmt(
            tablename,  # address_transactions_secondary_ids
            columns=["*"],
            where={
                k: "?"
                for k in [
                    id_group_col,
                ]
            },
            limit=1,
        )
        prep = self._db.get_prepared_statement(stmt)

        bstmts = [
            prep.bind({id_group_col: address_id_group})
            for address_id_group in address_id_groups
        ]

        return self._db.execute_statements_async(bstmts, concurrency=CONCURRENCY)

    def get_address_outgoing_relations_async_batch(
        self, rel_ids: List[Tuple[int, int]]
    ):
        stmt = self.select_stmt(
            "address_outgoing_relations",
            columns=["*"],
            where={
                k: "?"
                for k in ["src_address_id_group", "src_address_id", "dst_address_id"]
            },
            limit=1,
        )
        prep = self._db.get_prepared_statement(stmt)

        bstmts = [
            prep.bind(
                {
                    "src_address_id_group": self.get_id_group(
                        src_address, self.get_address_id_bucket_size()
                    ),
                    "src_address_id": src_address,
                    "dst_address_id": dst_address,
                }
            )
            for src_address, dst_address in rel_ids
        ]

        return self._db.execute_statements_async(bstmts, concurrency=CONCURRENCY)

    def get_address_outrelations_async_batch_account(
        self, rel_ids: List[Tuple[int, int]]
    ):
        stmt = self.select_stmt(
            "address_outgoing_relations",
            columns=["*"],
            where={
                k: "?"
                for k in [
                    "src_address_id_group",
                    "src_address_id_secondary_group",
                    "src_address_id",
                    "dst_address_id",
                ]
            },
            limit=1,
        )
        prep = self._db.get_prepared_statement(stmt)
        bucketsize = self.get_address_id_bucket_size()
        relations_nbuckets = self.get_addressrelations_ids_nbuckets()

        bstmts = []
        for src_address, dst_address in rel_ids:
            address_group, secondary_group = get_id_group_with_secondary_relations(
                src_address, dst_address, bucketsize, relations_nbuckets
            )
            bstmts.append(
                prep.bind(
                    {
                        "src_address_id_group": address_group,
                        "src_address_id_secondary_group": secondary_group,
                        "src_address_id": src_address,
                        "dst_address_id": dst_address,
                    }
                )
            )

        return self._db.execute_statements_async(bstmts, concurrency=CONCURRENCY)

    def get_balance_async_batch_account(self, address_ids: List[id]):
        stmt = self.select_stmt(
            "balance",
            columns=["*"],
            where={k: "?" for k in ["address_id_group", "address_id"]},
        )
        prep = self._db.get_prepared_statement(stmt)

        bstmts = [
            prep.bind(
                {
                    "address_id_group": self.get_id_group(
                        address_id, self.get_address_id_bucket_size()
                    ),
                    "address_id": address_id,
                }
            )
            for address_id in address_ids
        ]

        return self._db.execute_statements_async(bstmts, concurrency=CONCURRENCY)

    def get_address_outgoing_relations_async(
        self, address_id: int, dst_address_id: Optional[int]
    ):
        w = {
            "src_address_id_group": self.get_id_group(
                address_id, self.get_address_id_bucket_size()
            ),
            "src_address_id": address_id,
        }
        if dst_address_id is not None:
            w["dst_address_id"] = dst_address_id
        return self.select_async(
            "address_outgoing_relations",
            columns=["*"],
            where=w,
            limit=1,
        )

    def get_cluster_async_batch(self, cluster_ids: List[int]):
        bs = self.get_cluster_id_bucket_size()
        stmt = self.select_stmt(
            "cluster",
            columns=["*"],
            where={k: "?" for k in ["cluster_id_group", "cluster_id"]},
            limit=1,
        )
        prep = self._db.get_prepared_statement(stmt)

        bstmts = [
            prep.bind(
                {
                    "cluster_id_group": self.get_id_group(clstr_id, bs),
                    "cluster_id": clstr_id,
                }
            )
            for clstr_id in cluster_ids
        ]

        return zip(
            cluster_ids,
            self._db.execute_statements_async(bstmts, concurrency=CONCURRENCY),
        )

    def get_cluster_async(self, cluster_id: int):
        bucket = self.get_id_group(cluster_id, self.get_cluster_id_bucket_size())
        w = {"cluster_id_group": bucket, "cluster_id": f"{cluster_id}"}
        return self.select_async(
            "cluster",
            columns=["*"],
            where=w,
            limit=1,
        )

    def get_cluster_incoming_relations_async_batch(
        self, rel_ids: List[Tuple[int, int]]
    ):
        stmt = self.select_stmt(
            "cluster_incoming_relations",
            columns=["*"],
            where={
                k: "?"
                for k in ["dst_cluster_id_group", "dst_cluster_id", "src_cluster_id"]
            },
            limit=1,
        )
        prep = self._db.get_prepared_statement(stmt)

        bstmts = [
            prep.bind(
                {
                    "dst_cluster_id_group": self.get_id_group(
                        dst_address, self.get_address_id_bucket_size()
                    ),
                    "dst_cluster_id": dst_address,
                    "src_cluster_id": src_address,
                }
            )
            for dst_address, src_address in rel_ids
        ]

        return self._db.execute_statements_async(bstmts, concurrency=CONCURRENCY)

    def get_cluster_incoming_relations_async(
        self, cluster_id: int, src_cluster_id: Optional[int]
    ):
        w = {
            "dst_cluster_id_group": self.get_id_group(
                cluster_id, self.get_cluster_id_bucket_size()
            ),
            "dst_cluster_id": cluster_id,
        }
        if src_cluster_id is not None:
            w["src_cluster_id"] = src_cluster_id
        return self.select_async(
            "cluster_incoming_relations",
            columns=["*"],
            where=w,
            limit=1,
        )

    def get_cluster_outgoing_relations_async_batch(
        self, rel_ids: List[Tuple[int, int]]
    ):
        stmt = self.select_stmt(
            "cluster_outgoing_relations",
            columns=["*"],
            where={
                k: "?"
                for k in ["src_cluster_id_group", "src_cluster_id", "dst_cluster_id"]
            },
            limit=1,
        )
        prep = self._db.get_prepared_statement(stmt)

        bstmts = [
            prep.bind(
                {
                    "src_cluster_id_group": self.get_id_group(
                        src_address, self.get_address_id_bucket_size()
                    ),
                    "src_cluster_id": src_address,
                    "dst_cluster_id": dst_address,
                }
            )
            for src_address, dst_address in rel_ids
        ]

        return self._db.execute_statements_async(bstmts, concurrency=CONCURRENCY)

    def get_cluster_outgoing_relations_async(
        self, cluster_id: int, dst_cluster_id: Optional[int]
    ):
        w = {
            "src_cluster_id_group": self.get_id_group(
                cluster_id, self.get_cluster_id_bucket_size()
            ),
            "src_cluster_id": cluster_id,
        }
        if dst_cluster_id is not None:
            w["dst_cluster_id"] = dst_cluster_id
        return self.select_async(
            "cluster_outgoing_relations",
            columns=["*"],
            where=w,
            limit=1,
        )

    def has_delta_updater_v1_tables(self) -> bool:
        return (
            self._db.has_table(self._keyspace, "delta_updater_state")
            or self._db.has_table(self._keyspace, "dirty_addresses")
            or self._db.has_table(self._keyspace, "new_addresses")
        )


class AnalyticsDb:
    """Unified analytics DB interface"""

    def __init__(
        self, raw: KeyspaceConfig, transformed: KeyspaceConfig, db: CassandraDb
    ):
        self._raw_config = raw
        self._transformed_config = transformed
        self._db = db
        self._raw = raw.db_type(raw, db)
        self._transformed = transformed.db_type(transformed, db)

    def __repr__(self):
        return (
            f"Raw: {self._raw_config.keyspace_name}, "
            f"Transformed: {self._transformed_config.keyspace_name}, "
            f"DB: {self._db}"
        )

    def db(self) -> CassandraDb:
        return self._db

    def open(self):  # noqa:
        self._db.connect()

    def close(self):
        self._db.close()

    def __enter__(self):
        self.open()
        return self

    def clone(self) -> "AnalyticsDb":
        db_cloned = self._db.clone()
        return AnalyticsDb(self._raw_config, self._transformed_config, db_cloned)

    def __exit__(self, exc_type, exc_value, tb):
        self.close()

    def by_ks_type(self, keyspace_type):
        if keyspace_type == "transformed":
            return self.transformed
        elif keyspace_type == "raw":
            return self.raw
        else:
            raise Exception(f"Unknown keyspace type choose from {keyspace_types}")

    @property
    def raw(self) -> RawDb:
        return self._raw

    @property
    def transformed(self) -> TransformedDb:
        return self._transformed



from graphsenselib.deltaupdate.update.account.accountlegacy import (
    UpdateStrategyAccountLegacy,
)
from graphsenselib.deltaupdate.update.generic import ApplicationStrategy
from graphsenselib.deltaupdate.update.utxo.utxolegacy import UpdateStrategyUtxoLegacy

from ...config import currency_to_schema_type
from ...config.config import DeltaUpdaterConfig
from ...db import AnalyticsDb
from .abstractupdater import AbstractUpdateStrategy
from .account import UpdateStrategyAccount
from .utxo import UpdateStrategyUtxo


class UpdaterFactory:
    def get_updater(
        self,
        du_config: DeltaUpdaterConfig,
        db: AnalyticsDb,
        version: int,
        write_new: bool,
        write_dirty: bool,
        pedantic: bool,
        write_batch: int,
        patch_mode: bool,
        forward_fill_rates: bool = False,
    ) -> AbstractUpdateStrategy:
        currency = du_config.currency
        schema_type = currency_to_schema_type[currency]
        if schema_type == "utxo" and version == 1:
            return UpdateStrategyUtxoLegacy(db, currency, write_new, write_dirty)
        if schema_type == "utxo" and version == 2:
            app_strat = (
                ApplicationStrategy.BATCH if write_batch > 1 else ApplicationStrategy.TX
            )
            return UpdateStrategyUtxo(
                db,
                currency,
                pedantic,
                app_strat,
                patch_mode,
                forward_fill_rates=forward_fill_rates,
            )
        if (schema_type == "account" or schema_type == "account_trx") and version == 1:
            return UpdateStrategyAccountLegacy(
                db,
                currency,
                write_new,
                write_dirty,
                forward_fill_rates=forward_fill_rates,
            )
        if (schema_type == "account" or schema_type == "account_trx") and version == 2:
            app_strat = ApplicationStrategy.BATCH
            return UpdateStrategyAccount(
                db,
                du_config,
                pedantic,
                app_strat,
                patch_mode,
                forward_fill_rates=forward_fill_rates,
            )
        else:
            raise Exception(f"Unsupported schema type {schema_type} or {version}")



import json
import logging
import os

from .abstract import WatchConfig, WatcherState, WatchpointProvider

logger = logging.getLogger(__name__)


class JsonWatchpointProvider(WatchpointProvider):
    def __init__(self, filename: str):
        with open(filename) as f:
            self.data = json.load(f)

    def is_watched(self, address: str):
        return address in self.data

    def get_configuration(self, address: str) -> WatchConfig:
        d = self.data.get(address, None)
        return (
            WatchConfig(
                email=d["email"],
                on_outgoing=d["on_outgoing"],
                on_incoming=d["on_incoming"],
                value_gt=d.get("value_lgt", None),
            )
            if d
            else None
        )


class JsonWatcherState(WatcherState):
    def __init__(self, filename: str):
        if not os.path.isfile(filename):
            logging.info(
                "Watcher state file not found. " f"Creating new file {filename}"
            )
            with open(filename, "w") as f:
                json.dump({"block": 0}, f)

        self.filename = filename
        self.data = None

    def load(self):
        with open(self.filename) as f:
            self.data = json.load(f)

    def get_next_watch_block(self) -> int:
        return self.data.get("block", 0)

    def done_with_block(self):
        self.data["block"] = self.get_next_watch_block() + 1

    def persist(self):
        with open(self.filename, "w") as f:
            json.dump(self.data, f)



import bisect
import itertools
from datetime import timedelta
from typing import Iterable, Optional, Sequence

import pandas as pd


class DataObject:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def as_dict(self):
        return self.__dict__

    def __repr__(self):
        keys = sorted(self.__dict__)
        items = ("{}={!r}".format(k, self.__dict__[k]) for k in keys)
        return "{}({})".format(type(self).__name__, ", ".join(items))

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


def dict_to_dataobject(d: dict) -> DataObject:
    if type(d) == dict:
        obj = DataObject(**d)
        for k, v in obj.__dict__.items():
            obj.__dict__[k] = dict_to_dataobject(v)
        return obj
    if type(d) == list:
        return [dict_to_dataobject(x) for x in d]
    else:
        return d


class GenericArrayFacade:
    def __init__(self, getter_fun):
        self.getter = getter_fun

    def __getitem__(self, key):
        return self.getter(key)


def remove_prefix(s: str, prefix: str) -> str:
    if s.startswith(prefix):
        return s[len(prefix) :]
    else:
        return s


def group_by(lst: Iterable, key) -> dict:
    groups = itertools.groupby(sorted(lst, key=key), key=key)
    return {k: list(v) for k, v in groups}


def groupby_property(
    lst: Iterable, property_name: str, sort_by: Optional[str] = None
) -> dict:
    def keyfun(x):
        return getattr(x, property_name)

    groups = group_by(lst, key=keyfun)

    return {
        k: (v if sort_by is None else sorted(v, key=lambda x: getattr(x, sort_by)))
        for k, v in groups.items()
    }


def camel_to_snake_case(camel_str: str) -> str:
    return "".join([f"_{c.lower()}" if c.isupper() else c for c in camel_str])


def dict_with_snake_keys(d) -> dict:
    if type(d) != dict:
        return d
    return {camel_to_snake_case(k): dict_with_snake_keys(v) for k, v in d.items()}


def get_cassandra_result_as_dateframe(result):
    df = pd.DataFrame(result)
    return df


def fallback(*args):
    for arg in args:
        if arg is not None:
            return arg


def split_list_on_condition(lst, condition):
    return [x for x in lst if condition(x)], [x for x in lst if not condition(x)]


def remove_multi_whitespace(string):
    return " ".join(string.split())


def no_nones(lst):
    return [item for item in lst if item is not None]


def flatten(list_of_lists):
    return [item for sublist in list_of_lists for item in sublist]


def batch(iterable, n=1):
    length = len(iterable)
    for ndx in range(0, length, n):
        yield iterable[ndx : min(ndx + n, length)]


def binary_search(L, x, lo, hi):
    i = bisect.bisect_left(L, x, lo=lo, hi=hi)
    if i == hi or L[i] != x:
        return -1
    return i


def pandas_row_factory(colnames, rows):
    return pd.DataFrame(rows, columns=colnames)


def subkey_exists(item, key_list) -> bool:
    if item is None and len(key_list) > 0:
        return False
    if not key_list:
        return True
    h, *rest = key_list
    if h in item:
        return subkey_exists(item[h], rest)
    else:
        return False


def subkey_get(item, key_list) -> bool:
    if item is None and len(key_list) > 0:
        return item
    if not key_list:
        return item
    h, *rest = key_list
    if h.isdigit() and isinstance(item, list) and int(h) < len(item):
        return subkey_get(item[int(h)], rest)
    elif h in item:
        return subkey_get(item[h], rest)
    else:
        return None


def first_or_default(seq: Sequence[object], pred, default=None):
    return next(filter(pred, seq), default)


def batch_date(da, db, **kwargs):
    step = timedelta(**kwargs)
    while da <= db:
        start_old = da
        da += step
        if da > db:
            da = db
        yield (start_old, da)
        da += timedelta(days=1)


def generate_date_range_days(da, db):
    return itertools.takewhile(
        lambda x: x <= db,
        (da + timedelta(days=delta) for delta in itertools.count(start=0, step=1)),
    )



# flake8: noqa
import pickle
from io import StringIO

import pandas as pd

from graphsenselib.deltaupdate.update.account.createdeltas import (
    get_sorted_unique_addresses,
)
from graphsenselib.deltaupdate.update.account.modelsraw import (
    AccountBlockAdapter,
    AccountLogAdapter,
    AccountTransactionAdapter,
    EthTraceAdapter,
    TrxTraceAdapter,
    TrxTransactionAdapter,
)
from graphsenselib.deltaupdate.update.account.tokens import ERC20Decoder, TokenTransfer

data_eth = """
currency_ticker,assettype,decimals,token_address,peg_currency
USDT,ERC20,6,0xdac17f958d2ee523a2206206994597c13d831ec7,USD
USDC,ERC20,6,0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48,USD
WETH,ERC20,18,0xc02aaa39b223fe8d0a0e5c4f27ead9083c756cc2,ETH
"""

data_trx = """currency_ticker,assettype,decimals,token_address,coin_equivalent,peg_currency
USDT,TRC20,6,0xa614f803b6fd780986a42c78ec9c7f77e6ded13c,0,USD
USDC,TRC20,6,0x3487b63d30b5b2c87fb7ffa8bcfade38eaac1abe,0,USD
WTRX,TRC20,6,0x891cdb91d149f23b1a45d9c5ca78a88d0cb44c18,1,ETH
"""


currencies = ["trx", "eth"]
folder = "tests/deltaupdate/resources/account"
filetypes = ["transactions", "traces", "logs", "blocks"]

tx_schema = {
    "transaction_index": int,
    "tx_hash": bytes,
    "from_address": bytes,
    "to_address": bytes,
    "value": int,
    "gas_price": int,
    "transaction_type": int,
    "receipt_gas_used": int,
    "receipt_status": int,
    "block_id": int,
}

log_schema = {
    "block_id": int,
    "tx_hash": bytes,
    "log_index": int,
    "address": bytes,
    "topics": list,
    "data": bytes,
}
block_schema = {
    "block_id": int,
    "miner": bytes,
    "base_fee_per_gas": int,
    "gas_used": int,
}

trace_schema = {
    "block_id": int,
    "tx_hash": bytes,
    "trace_index": bytes,
    "from_address": bytes,
    "to_address": bytes,
    "value": int,
    "call_type": str,
    "status": int,
}


def load_data():
    # load txs, traces, logs, and blocks
    # read the jsons
    data = {
        currency: {
            filetype: pickle.load(open(f"{folder}/{currency}/{filetype}.pkl", "rb"))
            for filetype in filetypes
        }
        for currency in currencies
    }
    return data


def load_reference_data():
    lengths = {
        "trx": {
            "transactions": 293,
            "traces": 2,
            "logs": 76,
            "blocks": 1,
        },
        "eth": {
            "transactions": 117,
            "traces": 671,
            "logs": 300,
            "blocks": 1,
        },
    }

    blocks = {"trx": 50000011, "eth": 18000011}

    return lengths, blocks


def test_adapters_regression():
    data = load_data()
    lengths_ref, blocks_ref = load_reference_data()
    for currency in currencies:
        data_currency = data[currency]
        transactions, traces, logs, blocks = (
            data_currency["transactions"],
            data_currency["traces"],
            data_currency["logs"],
            data_currency["blocks"],
        )

        if currency == "trx":
            trace_adapter = TrxTraceAdapter()
            transaction_adapter = TrxTransactionAdapter()

        elif currency == "eth":
            trace_adapter = EthTraceAdapter()
            transaction_adapter = AccountTransactionAdapter()

        # convert dictionaries to dataclasses and unify naming
        log_adapter = AccountLogAdapter()
        block_adapter = AccountBlockAdapter()
        traces = trace_adapter.dicts_to_renamed_dataclasses(traces)
        traces = trace_adapter.process_fields_in_list(traces)
        transactions = transaction_adapter.dicts_to_dataclasses(transactions)
        logs = log_adapter.dicts_to_dataclasses(logs)
        blocks = block_adapter.dicts_to_dataclasses(blocks)

        length_ref = lengths_ref[currency]
        assert len(transactions) == length_ref["transactions"]
        assert len(traces) == length_ref["traces"]
        assert len(logs) == length_ref["logs"]
        assert len(blocks) == length_ref["blocks"]

        assert traces[0].block_id == blocks_ref[currency]

        # check that the files have the correct schema
        for transaction in transactions:
            assert isinstance(transaction, transaction_adapter.datamodel)
            assert isinstance(transaction.tx_hash, bytes)
            assert (
                isinstance(transaction.from_address, bytes)
                or transaction.from_address is None
            )
            assert isinstance(transaction.to_address, bytes)
            assert isinstance(transaction.value, int)
            assert isinstance(transaction.gas_price, int)
            assert isinstance(transaction.transaction_type, int)
            assert isinstance(transaction.receipt_gas_used, int)
            assert isinstance(transaction.receipt_status, int)
            assert isinstance(transaction.block_id, int)

        for trace in traces:
            assert isinstance(trace, trace_adapter.datamodel)
            assert isinstance(trace.tx_hash, bytes) or trace.tx_hash is None
            assert isinstance(trace.from_address, bytes) or trace.from_address is None
            assert isinstance(trace.to_address, bytes)
            assert isinstance(trace.value, int)
            assert isinstance(trace.call_type, str) or trace.call_type is None
            assert isinstance(trace.status, int)
            assert isinstance(trace.block_id, int)

        for log in logs:
            assert isinstance(log, log_adapter.datamodel)
            assert isinstance(log.block_id, int)
            assert isinstance(log.tx_hash, bytes)
            assert isinstance(log.log_index, int)
            assert isinstance(log.address, bytes)
            assert isinstance(log.topics, list)
            assert isinstance(log.data, bytes)

        for block in blocks:
            assert isinstance(block, block_adapter.datamodel)
            assert isinstance(block.block_id, int)
            assert isinstance(block.miner, bytes)
            assert isinstance(block.base_fee_per_gas, int)
            assert isinstance(block.gas_used, int)


def test_tokens_detected():
    SUPPORTED_TOKENS = pd.read_csv(StringIO(data_eth))

    tokendecoder = ERC20Decoder("eth", SUPPORTED_TOKENS)
    data = load_data()
    eth_data = data["eth"]
    logs = eth_data["logs"]
    log_adapter = AccountLogAdapter()
    logs = log_adapter.dicts_to_dataclasses(logs)
    token_transfers = [tokendecoder.log_to_transfer(log) for log in logs]
    token_transfers = [x for x in token_transfers if x is not None]

    assert len(token_transfers) == 51


def test_token_decoding():
    adapter = AccountLogAdapter()
    # todo move to tests
    example_log = {
        "log_index": 0,
        "transaction_index": 1,
        "block_hash": b"\x00\x00\x00\x00\x02\xfa\xf0\xe5A\xeab\x1d\xed\xc7%\x00\x074^"
        b"\x10\xaa5\xe7\xbd\xb7\xa9\x1c\xee\x99\x0f96",
        "address": b"\xa6\x14\xf8\x03\xb6\xfdx\t\x86\xa4,x\xec\x9c\x7fw\xe6\xde\xd1<",
        "data": b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\xba\x81@",
        "topics": [
            b"\xdd\xf2R\xad\x1b\xe2\xc8\x9bi\xc2\xb0h\xfc7\x8d\xaa\x95+\xa7\xf1c"
            b"\xc4\xa1\x16(\xf5ZM\xf5#\xb3\xef",
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb3\xa8"da\xf0\xe6'
            b"\xa9\xa1\x06?\xeb\xea\x88\xc6\xf6\xa5\xa0\x85~",
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xf0LT\xf6\xb6\xa2"
            b'\x9a\xf0ZT\x95\x8cIt\xc3\x83\xb4\xd9"\xac',
        ],
        "tx_hash": b"\xe0}\xe10\xe5\xc2\xb1\xde\x13\xcd\x88\xee!\xfa\x1e\xca]e\xba\xbb"
        b"\xecVG\xd3\x1c\xb7\x90\x1f\xc92wo",
        "block_id": 50000101,
        "block_id_group": 50000,
        "partition": 500,
        "topic0": b"\xdd\xf2R\xad\x1b\xe2\xc8\x9bi\xc2\xb0h\xfc7\x8d\xaa\x95+\xa7\xf1c"
        b"\xc4\xa1\x16(\xf5ZM\xf5#\xb3\xef",
    }

    example_log = adapter.dict_to_dataclass(example_log)
    SUPPORTED_TOKENS = pd.read_csv(StringIO(data_eth))

    decoder = ERC20Decoder("eth", SUPPORTED_TOKENS)
    decoded_transfer = decoder.log_to_transfer(example_log)
    assert decoded_transfer is None

    example_log = {
        "log_index": 0,
        "transaction_index": 1,
        "block_hash": b"\x00\x00\x00\x00\x02\xfa\xf0\xe5A\xeab\x1d\xed\xc7%\x00\x074^"
        b"\x10\xaa5\xe7\xbd\xb7\xa9\x1c\xee\x99\x0f96",
        "address": bytes.fromhex("dac17f958d2ee523a2206206994597c13d831ec7"),
        "data": b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\xba\x81@",
        "topics": [
            b"\xdd\xf2R\xad\x1b\xe2\xc8\x9bi\xc2\xb0h"
            b"\xfc7\x8d\xaa\x95+\xa7\xf1c\xc4\xa1"
            b"\x16(\xf5ZM\xf5#\xb3\xef",
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb3\xa8"da\xf0\xe6\xa9'
            b"\xa1\x06?\xeb\xea\x88\xc6\xf6\xa5\xa0\x85~",
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xf0LT\xf6\xb6\xa2\x9a"
            b'\xf0ZT\x95\x8cIt\xc3\x83\xb4\xd9"\xac',
        ],
        "tx_hash": b"\xe0}\xe10\xe5\xc2\xb1\xde\x13\xcd\x88\xee!\xfa\x1e\xca]e\xba\xbb"
        b"\xecVG\xd3\x1c\xb7\x90\x1f\xc92wo",
        "block_id": 50000101,
        "block_id_group": 50000,
        "partition": 500,
        "topic0": b"\xdd\xf2R\xad\x1b\xe2\xc8\x9bi\xc2\xb0h\xfc7\x8d\xaa\x95+\xa7\xf1c"
        b"\xc4\xa1\x16(\xf5ZM\xf5#\xb3\xef",
    }

    example_log = adapter.dict_to_dataclass(example_log)
    decoder = ERC20Decoder("eth", SUPPORTED_TOKENS)
    decoded_transfer = decoder.log_to_transfer(example_log)
    check = TokenTransfer(
        from_address=bytes.fromhex("B3a8226461F0e6A9a1063fEBeA88C6f6A5a0857E"),
        to_address=bytes.fromhex("F04C54F6b6A29aF05A54958c4974C383B4D922ac"),
        value=29000000,
        asset="USDT",
        coin_equivalent=0,
        usd_equivalent=1,
        block_id=50000101,
        tx_hash=b"\xe0}\xe10\xe5\xc2\xb1\xde\x13\xcd\x88\xee!\xfa\x1e\xca]e\xba\xbb"
        b"\xecVG\xd3\x1c\xb7\x90\x1f\xc92wo",
        log_index=0,
        decimals=6,
    )

    assert decoded_transfer == check

    SUPPORTED_TOKENS = pd.read_csv(StringIO(data_trx))
    decoder = ERC20Decoder("trx", SUPPORTED_TOKENS)
    decoded_transfer = decoder.log_to_transfer(example_log)
    assert decoded_transfer is None


def test_address_sorting():
    class SortableAssetTransfer:
        def __init__(
            self,
            from_address=None,
            to_address=None,
            block_id=None,
            log_index=None,
            trace_index=None,
            transaction_index=None,
        ):
            self.from_address = from_address
            self.to_address = to_address
            self.block_id = block_id
            self.log_index = log_index
            self.trace_index = trace_index
            self.transaction_index = transaction_index

    traces_s = [
        SortableAssetTransfer(
            from_address="0x1", to_address="0x2", block_id=1, trace_index=2
        ),
        SortableAssetTransfer(
            from_address="0x3", to_address="0x4", block_id=2, trace_index=1
        ),
    ]
    reward_traces = [SortableAssetTransfer(to_address="0x5", block_id=3, trace_index=0)]
    token_transfers = [
        SortableAssetTransfer(
            from_address="0x2", to_address="0x3", block_id=1, log_index=1
        ),
        SortableAssetTransfer(
            from_address="0x4", to_address="0x1", block_id=2, log_index=2
        ),
    ]
    transactions = [
        SortableAssetTransfer(
            from_address="0x1", to_address="0x0", block_id=1, transaction_index=1000001
        ),
        SortableAssetTransfer(
            from_address="0x2", to_address="0x4", block_id=2, transaction_index=1000002
        ),
    ]

    expected_addresses = ["0x0", "0x1", "0x2", "0x3", "0x4", "0x5"]

    result_addresses = list(
        get_sorted_unique_addresses(
            traces_s, reward_traces, token_transfers, transactions
        )
    )

    assert result_addresses == expected_addresses



import logging
import time
from datetime import datetime
from typing import Any, Callable, Dict, List, NamedTuple, Optional, Set, Tuple

from cassandra import InvalidRequest

from graphsenselib.datatypes import DbChangeType, EntityType
from graphsenselib.db import AnalyticsDb, DbChange
from graphsenselib.deltaupdate.update.abstractupdater import (
    TABLE_NAME_DELTA_HISTORY,
    UpdateStrategy,
)
from graphsenselib.deltaupdate.update.generic import (
    Action,
    ApplicationStrategy,
    DbDelta,
    DeltaValue,
    EntityDelta,
    RawEntityTx,
    RelationDelta,
    prepare_entities_for_ingest,
    prepare_relations_for_ingest,
    prepare_txs_for_ingest,
)
from graphsenselib.rates import convert_to_fiat
from graphsenselib.utils import DataObject as MutableNamedTuple
from graphsenselib.utils import group_by, no_nones
from graphsenselib.utils.errorhandling import CrashRecoverer
from graphsenselib.utils.logging import LoggerScope
from graphsenselib.utils.utxo import (
    get_regflow,
    get_total_input_sum,
    get_unique_addresses_from_transaction,
    get_unique_addresses_from_transactions,
    get_unique_ordered_input_addresses_from_transactions,
    get_unique_ordered_output_addresses_from_transactions,
    regularize_inoutputs,
)

logger = logging.getLogger(__name__)


COINBASE_PSEUDO_ADDRESS = "coinbase"
PSEUDO_ADDRESS_AND_IDS = {COINBASE_PSEUDO_ADDRESS: 0}

DEFAULT_SUMMARY_STATISTICS = MutableNamedTuple(
    **{
        "id": 0,
        "timestamp": 0,
        "timestamp_transform": 0,
        "no_blocks": 0,
        "no_blocks_transform": 0,
        "no_transactions": 0,
        "no_addresses": 0,
        "no_address_relations": 0,
        "no_clusters": 0,
        "no_cluster_relations": 0,
    }
)


def get_table_abbrev(table_name: str) -> str:
    return "".join([f"{x[:1]}" for x in table_name.split("_")])


def dbdelta_from_utxo_transaction(tx: dict, rates: List[int]) -> DbDelta:
    """Create a DbDelta instance form a transaction

    Args:
        tx (dict): transaction to build the delta from
        rates (List[int]): convertion rates to use.

    Returns:
        DbDelta: delta to apply to the db.
    """
    tx_adrs = get_unique_addresses_from_transaction(tx)
    reg_in = regularize_inoutputs(tx.inputs)
    reg_out = regularize_inoutputs(tx.outputs)

    reginput_sum = sum([v for _, v in reg_in.items()])
    flows = {adr: get_regflow(reg_in, reg_out, adr) for adr in tx_adrs}
    input_flows_sum = sum([f for adr, f in flows.items() if adr in reg_in and f <= 0])
    """
        reginput_sum == -input_flows_sum,
        unless input address is used as output in same tx
    """
    reduced_input_sum = get_total_input_sum(tx.inputs) - (
        reginput_sum + input_flows_sum
    )

    entity_updates = []
    new_entity_transactions = []
    relations_updates = []

    for adr, value in reg_in.items():
        entity_updates.append(
            EntityDelta(
                identifier=adr,
                total_spent=DeltaValue(
                    value=value, fiat_values=convert_to_fiat(value, rates)
                ),
                total_received=DeltaValue(
                    value=0, fiat_values=convert_to_fiat(0, rates)
                ),
                first_tx_id=tx.tx_id,
                last_tx_id=tx.tx_id,
                no_incoming_txs=0,
                no_outgoing_txs=1,
            )
        )

    for adr, value in reg_out.items():
        entity_updates.append(
            EntityDelta(
                identifier=adr,
                total_spent=DeltaValue(value=0, fiat_values=convert_to_fiat(0, rates)),
                total_received=DeltaValue(
                    value=value, fiat_values=convert_to_fiat(value, rates)
                ),
                first_tx_id=tx.tx_id,
                last_tx_id=tx.tx_id,
                no_incoming_txs=1,
                no_outgoing_txs=0,
            )
        )

    for adr in tx_adrs:
        flow = flows[adr]
        new_entity_transactions.append(
            RawEntityTx(
                identifier=adr, is_outgoing=(flow < 0), value=flow, tx_id=tx.tx_id
            )
        )

    for iadr, _ in reg_in.items():
        for oadr, _ in reg_out.items():
            if iadr == oadr:
                continue

            iflow = flows[iadr]
            oflow = flows[oadr]
            if reduced_input_sum == 0:
                # This can happen for txs with zero value and
                # zero fees. eg. in btc
                # c1e0db6368a43f5589352ed44aa1ff9af33410e4a9fd9be0f6ac42d9e4117151
                v = 0
            else:
                v = abs(round((iflow / reduced_input_sum) * oflow))
            assert v <= max(abs(iflow), abs(oflow))
            relations_updates.append(
                RelationDelta(
                    src_identifier=iadr,
                    dst_identifier=oadr,
                    estimated_value=DeltaValue(
                        value=v, fiat_values=convert_to_fiat(v, rates)
                    ),
                    no_transactions=1,
                )
            )

    return DbDelta(
        entity_updates=entity_updates,
        new_entity_txs=new_entity_transactions,
        relation_updates=relations_updates,
    )


def get_transaction_changes(
    db: AnalyticsDb,
    txs: List,
    rates: Dict[int, List],
    get_next_address_id: Callable[[], int],
    get_next_cluster_id: Callable[[], int],
) -> Tuple[List[DbChange], int, int, int, int]:
    """Main function to transform a list of transactions from the raw
    keyspace to changes to the transformed db.

    Args:
        db (AnalyticsDb): database instance
        txs (List): list of transaction rows (raw db schema)
        rates (Dict[int, List]): Mapping from block to exchange rates
        (transformed db schema)
        get_next_address_id (Callable[[], int]): Function to fetch next new address_id
        get_next_cluster_id (Callable[[], int]): Function to fetch next new cluster_id
    """
    tdb = db.transformed

    """
        Add pseudo inputs for coinbase txs
    """
    new_txs = []

    for tx in txs:
        if tx.coinbase:
            assert tx.inputs is None
            outputsum = (
                sum([o.value for o in tx.outputs]) if tx.outputs is not None else 0
            )
            new_txs.append(
                tx._replace(
                    inputs=[
                        MutableNamedTuple(
                            **{
                                "address": [COINBASE_PSEUDO_ADDRESS],
                                "value": outputsum,
                                "address_type": None,
                            }
                        )
                    ]
                )
            )
        else:
            # non-coinbase tx, nothing to do
            new_txs.append(tx)

    txs = new_txs
    del new_txs

    """
        Build dict of unique addresses in the batch.
    """
    addresses = get_unique_addresses_from_transactions(txs)

    len_addr = len(addresses)

    """
        Start loading the address_ids for the addresses async
    """
    with LoggerScope.debug(logger, f"Checking existence for {len_addr} addresses") as _:
        addr_ids = {  # noqa: C416
            adr: address_id
            for adr, address_id in db.transformed.get_address_id_async_batch(
                list(addresses)
            )
        }

        del addresses

    """
        Sort transactions by block and tx id.
        Database should already return them sorted.
        This is just for caution
    """
    with LoggerScope.debug(logger, "Prepare transaction data") as lg:
        txs = sorted(txs, key=lambda row: (row.block_id, row.tx_id))
        lg.debug(f"Working on batch of {len(txs)} transactions.")
        ordered_output_addresses = (
            get_unique_ordered_output_addresses_from_transactions(txs)
        )
        ordered_input_addresses = get_unique_ordered_input_addresses_from_transactions(
            txs
        )

    """
        Compute the changeset for each tx of the batch and convert
        all currency values with the corresponding rates.
    """
    with LoggerScope.debug(logger, "Creating address changeset") as lg:
        per_tx_changes = []

        for tx in txs:
            per_tx_changes.append(dbdelta_from_utxo_transaction(tx, rates[tx.block_id]))

        del rates
        del txs

        """
            Aggregate and compress the changeset to minimize database writes
        """
        address_delta = DbDelta.merge(per_tx_changes)
        del per_tx_changes

    """
        Read address data to merge for address updates
    """
    with LoggerScope.debug(logger, "Reading addresses to be updated") as lg:
        existing_addr_ids = no_nones(
            [address_id.result_or_exc.one() for adr, address_id in addr_ids.items()]
        )
        addresses_resolved = {  # noqa: C416
            addr_id: address
            for addr_id, address in tdb.get_address_async_batch(
                [adr.address_id for adr in existing_addr_ids]
            )
        }
        del existing_addr_ids

        def get_resolved_address(addr_id_exc):
            addr_id = addr_id_exc.result_or_exc.one()
            return (
                (None, None)
                if addr_id is None
                else (
                    addr_id.address_id,
                    addresses_resolved[addr_id.address_id].result_or_exc.one(),  # noqa
                )
            )

        addresses = {
            adr: get_resolved_address(address_id)
            for adr, address_id in addr_ids.items()
        }

        del addresses_resolved

    def get_next_address_ids_with_aliases(address: str):
        return (
            get_next_address_id()
            if address not in PSEUDO_ADDRESS_AND_IDS
            else PSEUDO_ADDRESS_AND_IDS[address]
        )

    with LoggerScope.debug(
        logger, "Assigning new address ids and cluster ids for new addresses"
    ) as lg:
        new_cluster_ids = {}

        for out_addr in ordered_output_addresses:
            addr_id, address = addresses[out_addr]
            if addr_id is None:
                new_addr_id = get_next_address_ids_with_aliases(address)
                addresses[out_addr] = (new_addr_id, None)
                new_cluster_ids[new_addr_id] = get_next_cluster_id()
            elif address is None:
                """
                I have found cases where address_prefix table is written
                But not the address table, for those cases we set a new
                cluster id and create the address in turn
                """
                lg.warning(
                    f"Address {out_addr} has address "
                    f"id {addr_id} but no address entry"
                )
                new_cluster_ids[addr_id] = get_next_cluster_id()

        del ordered_output_addresses

        not_yet_seen_input_addresses = {
            k for k, (addr_id2, _) in addresses.items() if addr_id2 is None
        }

        if len(not_yet_seen_input_addresses) > 0:
            for out_addr in ordered_input_addresses:
                if out_addr in not_yet_seen_input_addresses:
                    new_addr_id = get_next_address_ids_with_aliases(out_addr)
                    lg.debug(
                        "Encountered new input address - "
                        f"{out_addr:64}. Creating it with id {new_addr_id}."
                    )
                    addresses[out_addr] = (new_addr_id, None)
                    new_cluster_ids[new_addr_id] = get_next_cluster_id()

            del ordered_input_addresses

        del not_yet_seen_input_addresses

        assert (
            len([1 for k, (addr_id2, _) in addresses.items() if addr_id2 is None]) == 0
        )

    """
        Reading relations to be updated.
    """
    with LoggerScope.debug(logger, "Reading address relations to be updated") as _:
        rel_to_query = [
            (addresses[update.src_identifier][0], addresses[update.dst_identifier][0])
            for update in address_delta.relation_updates
        ]
        addr_outrelations_q = tdb.get_address_outgoing_relations_async_batch(
            rel_to_query
        )
        addr_outrelations = {
            (update.src_identifier, update.dst_identifier): qr
            for update, qr in zip(address_delta.relation_updates, addr_outrelations_q)
        }

        rel_to_query = [
            (addresses[update.dst_identifier][0], addresses[update.src_identifier][0])
            for update in address_delta.relation_updates
        ]
        addr_inrelations_q = tdb.get_address_incoming_relations_async_batch(
            rel_to_query
        )
        addr_inrelations = {
            (update.src_identifier, update.dst_identifier): qr
            for update, qr in zip(address_delta.relation_updates, addr_inrelations_q)
        }

        del rel_to_query, addr_inrelations_q, addr_outrelations_q

    with LoggerScope.debug(logger, "Reading clusters for addresses") as _:
        clusters_resolved = {  # noqa: C416
            cluster_id: cluster
            for cluster_id, cluster in tdb.get_cluster_async_batch(
                [
                    address.cluster_id
                    for adr, (addr_id, address) in addresses.items()
                    if address is not None
                ]
            )
        }

        def get_resolved_cluster(address_tuple):
            aidr, address = address_tuple
            if address is not None:
                assert address.cluster_id not in new_cluster_ids.values()
                assert aidr == address.address_id
                return (
                    aidr,
                    address,
                    address.cluster_id,
                    clusters_resolved[address.cluster_id].result_or_exc.one(),  # noqa
                )
            else:
                return (aidr, None, new_cluster_ids[aidr], None)

        """Assigning new address ids and cluster ids for new addresses"""

        addresses_with_cluster = {
            adr: get_resolved_cluster(address_tuple)
            for adr, address_tuple in addresses.items()
        }
        del clusters_resolved, addresses

    with LoggerScope.debug(logger, "Creating local lookup tables") as lg:
        cluster_to_addr_id = group_by(
            [
                (cluster_id, address_id)
                for _, (
                    address_id,
                    _,
                    cluster_id,
                    _,
                ) in addresses_with_cluster.items()
            ],
            lambda x: x[0],
        )

        cluster_from_cluster_id = group_by(
            [
                (cluster_id, cluster)
                for _, (_, _, cluster_id, cluster) in addresses_with_cluster.items()
            ],
            lambda x: x[0],
        )

        def address_to_cluster_id(addr: str) -> int:
            return addresses_with_cluster[addr][2]

        def address_to_address_id(addr: str) -> int:
            return addresses_with_cluster[addr][0]

        def address_to_address_obj(addr: str) -> int:
            return addresses_with_cluster[addr][1]

        def address_to_cluster(addr: str) -> int:
            return addresses_with_cluster[addr][3]

        def cluster_id_to_address_id(cluster_id: int) -> Set[int]:
            s = {addr_id for _, addr_id in cluster_to_addr_id[cluster_id]}
            if len(s) != 1:
                raise Exception(f"Found more than one address id for cluster_id {s}")
            return list(s)[0]

        def cluster_id_to_cluster(cluster_id: int) -> Any:
            clusters = [
                x for (_, x) in cluster_from_cluster_id[cluster_id] if x is not None
            ]
            assert len({x.cluster_id for x in clusters}) <= 1
            return clusters[0] if len(clusters) > 0 else None

        def get_address_prefix(address_str):
            address = tdb.to_db_address(address_str)
            return (address.db_encoding, address.prefix)

    with LoggerScope.debug(logger, "Creating cluster changeset") as lg:
        cluster_delta = address_delta.to_cluster_delta(address_to_cluster_id)

    with LoggerScope.debug(logger, "Reading cluster relations to be updated") as lg:
        rel_to_query = [
            (update.src_identifier, update.dst_identifier)
            for update in cluster_delta.relation_updates
        ]
        clstr_outrelations_q = tdb.get_cluster_outgoing_relations_async_batch(
            rel_to_query
        )
        clstr_outrelations = {
            (update.src_identifier, update.dst_identifier): qr
            for update, qr in zip(cluster_delta.relation_updates, clstr_outrelations_q)
        }

        rel_to_query = [
            (update.dst_identifier, update.src_identifier)
            for update in cluster_delta.relation_updates
        ]
        clstr_inrelations_q = tdb.get_cluster_incoming_relations_async_batch(
            rel_to_query
        )
        clstr_inrelations = {
            (update.src_identifier, update.dst_identifier): qr
            for update, qr in zip(cluster_delta.relation_updates, clstr_inrelations_q)
        }

        del rel_to_query, clstr_inrelations_q, clstr_outrelations_q

    """
        Merge Db Entries with deltas
    """

    with LoggerScope.debug(logger, "Preparing data to be written.") as lg:
        changes = []

        ingest_configs = {
            EntityType.ADDRESS: {
                "bucket_size": tdb.get_address_id_bucket_size(),
                "delta": address_delta,
                "id_transformation": address_to_address_id,
                "get_entity": address_to_address_obj,
                "incoming_relations_db": addr_inrelations,
                "outgoing_relations_db": addr_outrelations,
            },
            EntityType.CLUSTER: {
                "bucket_size": tdb.get_cluster_id_bucket_size(),
                "delta": cluster_delta,
                "id_transformation": (lambda x: x),
                "get_entity": cluster_id_to_cluster,
                "incoming_relations_db": clstr_inrelations,
                "outgoing_relations_db": clstr_outrelations,
            },
        }

        new_relations_in = {}
        new_relations_out = {}
        nr_new_relations = {}
        nr_new_entities_created = {}
        for mode, config in ingest_configs.items():
            lg.debug(f"Prepare {mode} data.")
            """
            Creating new address/cluster transaction
            """
            changes.extend(
                prepare_txs_for_ingest(
                    config["delta"].new_entity_txs,
                    config["id_transformation"],
                    config["bucket_size"],
                    mode=mode,
                )
            )

            """ Merging relations deltas """
            (
                changes_relations,
                new_rels_in,
                new_rels_out,
                nr_new_rel_total,
            ) = prepare_relations_for_ingest(
                config["delta"].relation_updates,
                config["id_transformation"],
                config["incoming_relations_db"],
                config["outgoing_relations_db"],
                config["bucket_size"],
                mode=mode,
            )
            nr_new_rels = sum(new_rels_in.values()) + sum(new_rels_out.values())
            assert nr_new_rels == nr_new_rel_total
            nr_new_relations[mode] = nr_new_rels

            new_relations_in[mode] = new_rels_in
            new_relations_out[mode] = new_rels_out
            changes.extend(changes_relations)

            """ Merging entity deltas """
            entity_changes, nr_new_entities = prepare_entities_for_ingest(
                config["delta"].entity_updates,
                config["id_transformation"],
                config["get_entity"],
                address_to_cluster_id,
                cluster_id_to_address_id,
                new_relations_in[mode],
                new_relations_out[mode],
                config["bucket_size"],
                get_address_prefix,
                mode=mode,
            )
            changes.extend(entity_changes)
            nr_new_entities_created[mode] = nr_new_entities
            del nr_new_entities

    return (
        changes,
        nr_new_entities_created[EntityType.ADDRESS],
        nr_new_entities_created[EntityType.CLUSTER],
        nr_new_relations[EntityType.ADDRESS],
        nr_new_relations[EntityType.CLUSTER],
    )


def get_bookkeeping_changes(
    base_statistics: MutableNamedTuple,
    current_statistics: NamedTuple,
    last_block_processed: int,
    nr_new_address_relations: int,
    nr_new_addresses: int,
    nr_new_cluster_relations: int,
    nr_new_clusters: int,
    nr_new_tx: int,
    highest_address_id: int,
    runtime_seconds: int,
    bts: Dict[int, datetime],
    patch_mode: bool,
) -> List[DbChange]:
    """Creates changes for the bookkeeping tables like summary statistics after
    other data has been updated.

    Args:
        base_statistics (MutableNamedTuple): statistics db row, all the other
        parameters are note data is updated in this process
        current_statistics (NamedTuple): Current value of db statistics for comparison
        last_block_processed (int): Last block processed
        nr_new_address_relations (int): Delta new addresses relations in changeset
        nr_new_addresses (int): Delta new addresses in changeset
        nr_new_cluster_relations (int): Delta new cluster relations in changeset
        nr_new_clusters (int): Delta new clusters in changeset
        nr_new_tx (int): Delta new txs in changeset
        highest_address_id (int): current highest address_id
        runtime_seconds (int): runtime to create the last changes in seconds
        bts (Dict[int, datetime]): mapping from block to its timestamp
        delta values
    """
    changes = []
    with LoggerScope.debug(logger, "Creating summary_statistics updates") as lg:
        lb_date = bts[last_block_processed]
        stats = base_statistics
        no_blocks = last_block_processed + 1

        """ Update local stats """
        if not patch_mode:
            """when in patch mode (end block set by user)"""
            stats.no_blocks = no_blocks
        stats.timestamp = int(lb_date.timestamp())
        stats.no_address_relations += nr_new_address_relations
        stats.no_addresses += nr_new_addresses
        stats.no_cluster_relations += nr_new_cluster_relations
        stats.no_clusters += nr_new_clusters
        stats.no_transactions += nr_new_tx

        statistics = stats.as_dict()
        if current_statistics is not None and current_statistics.no_blocks != no_blocks:
            if not patch_mode:
                assert current_statistics.no_blocks < no_blocks
            # changes.append(
            #     DbChange.delete(
            #         table="summary_statistics",
            #         data={"no_blocks": current_statistics.no_blocks},
            #     )
            # )
        changes.append(DbChange.new(table="summary_statistics", data=statistics))

        lg.debug(f"Statistics: {statistics}")

        data_history = {
            "last_synced_block": last_block_processed,
            "last_synced_block_timestamp": lb_date,
            "highest_address_id": highest_address_id,
            "timestamp": datetime.now(),
            "write_new": False,
            "write_dirty": False,
            "runtime_seconds": runtime_seconds,
        }
        changes.append(DbChange.new(table=TABLE_NAME_DELTA_HISTORY, data=data_history))

        lg.debug(f"History: {data_history}")
    return changes


def validate_changes(db: AnalyticsDb, changes: List[DbChange]):
    """Validate a set of changes for correctness and consistency

    Args:
        db (AnalyticsDb): database instance
        changes (List[DbChange]): list of changes to be validated

    Raises:
        AssertionExecption: In case the data is in-correct.
    """
    with LoggerScope.debug(logger, "Validating changes (pedantic mode)") as _:
        tdb = db.transformed
        addresses_seen = {}
        cluster_seen = {}
        cluster_new = {}
        addresses_new = {}
        current_summary_stats = (
            db.transformed.get_summary_statistics() or DEFAULT_SUMMARY_STATISTICS
        )
        for change in changes:
            if change.action == DbChangeType.NEW and change.table == "cluster":
                # only one update per batch
                if change.data["cluster_id"] in cluster_seen:
                    raise ValueError(f"Only one update per cluster allowed: {change}")
                cluster_seen[change.data["cluster_id"]] = True
                cluster_new[change.data["cluster_id"]] = True

                if not (
                    len(list(tdb.get_cluster_async(change.data["cluster_id"]).result()))
                    == 0
                ):
                    raise ValueError(f"New cluster id is already in db: {change}")

            elif change.action == DbChangeType.NEW and change.table == "address":
                # only one update per batch
                if change.data["address_id"] in addresses_seen:
                    raise ValueError(f"Only one update per address allowed: {change}")
                addresses_seen[change.data["address_id"]] = True
                addresses_new[change.data["address_id"]] = True

                if not ("cluster_id" in change.data and change.data["cluster_id"] > 0):
                    raise ValueError(f"No cluster id in new address: {change}")
                if not (
                    len(list(tdb.get_address_async(change.data["address_id"]).result()))
                    == 0
                ):
                    raise ValueError(f"New address_id already in database: {change}")

            elif change.action == DbChangeType.UPDATE and change.table == "cluster":
                # only one update per batch
                if change.data["cluster_id"] in cluster_seen:
                    raise ValueError(f"Only one update per cluster allowed: {change}")
                cluster_seen[change.data["cluster_id"]] = True

                if "address_id" in change.data:
                    raise ValueError(
                        "Found address id in cluster update. "
                        f"Confused address with cluster? {change}"
                    )

                clusters = list(
                    tdb.get_cluster_async(change.data["cluster_id"]).result()
                )
                if not (len(clusters) == 1):
                    raise ValueError(
                        f"Could not find cluster to update in db! {change}"
                    )

                cluster = clusters[0]

                if cluster.cluster_id != change.data["cluster_id"]:
                    raise ValueError(f"Cluster_id do not match {cluster} {change}")

                if cluster.no_incoming_txs > change.data["no_incoming_txs"]:
                    raise ValueError(
                        f"Cluster changes are inconsistent {cluster} {change}"
                    )

                if cluster.no_outgoing_txs > change.data["no_outgoing_txs"]:
                    raise ValueError(
                        f"Cluster changes are inconsistent {cluster} {change}"
                    )

                if cluster.first_tx_id != change.data["first_tx_id"]:
                    raise ValueError(
                        f"Cluster changes are inconsistent {cluster} {change}"
                    )

                if cluster.last_tx_id >= change.data["last_tx_id"]:
                    raise ValueError(
                        f"Cluster changes are inconsistent {cluster} {change}"
                    )

                if cluster.total_received.value > change.data["total_received"].value:
                    raise ValueError(
                        f"Cluster changes are inconsistent {cluster} {change}"
                    )

                if cluster.total_spent.value > change.data["total_spent"].value:
                    raise ValueError(
                        f"Cluster changes are inconsistent {cluster} {change}"
                    )

                if cluster.in_degree > change.data["in_degree"]:
                    raise ValueError(
                        f"Cluster changes are inconsistent {cluster} {change}"
                    )

                if cluster.out_degree > change.data["out_degree"]:
                    raise ValueError(
                        f"Cluster changes are inconsistent {cluster} {change}"
                    )

                del cluster

            elif change.action == DbChangeType.UPDATE and change.table == "address":
                # only one update per batch
                if change.data["address_id"] in addresses_seen:
                    raise ValueError(f"Only one update per address allowed. {change}")
                addresses_seen[change.data["address_id"]] = True

                if "cluster_id" in change.data:
                    raise ValueError(
                        "Found cluster id in address update. "
                        f"Confused address with cluster? {change}"
                    )

                ad = tdb.get_address_async(change.data["address_id"]).result().one()
                if ad is None:
                    raise ValueError(f"Did not find address_id to update! {change}")

                if ad.no_incoming_txs > change.data["no_incoming_txs"]:
                    raise ValueError(f"Address changes are inconsistent {ad} {change}")

                if ad.no_outgoing_txs > change.data["no_outgoing_txs"]:
                    raise ValueError(f"Address changes are inconsistent {ad} {change}")

                if ad.first_tx_id != change.data["first_tx_id"]:
                    raise ValueError(f"Address changes are inconsistent {ad} {change}")

                if ad.last_tx_id >= change.data["last_tx_id"]:
                    raise ValueError(f"Address changes are inconsistent {ad} {change}")

                if ad.total_received.value > change.data["total_received"].value:
                    raise ValueError(f"Address changes are inconsistent {ad} {change}")

                if ad.total_spent.value > change.data["total_spent"].value:
                    raise ValueError(f"Address changes are inconsistent {ad} {change}")

                if ad.in_degree > change.data["in_degree"]:
                    raise ValueError(f"Address changes are inconsistent {ad} {change}")

                if ad.out_degree > change.data["out_degree"]:
                    raise ValueError(f"Address changes are inconsistent {ad} {change}")

                adid = tdb.get_address_id_async(ad.address).result().one()
                if adid is None:
                    raise ValueError(f"Did not find address to update! {change}")

                del ad, adid
            elif (
                change.action == DbChangeType.UPDATE
                and change.table == "address_incoming_relations"
            ):
                if not (
                    len(
                        list(
                            tdb.get_address_incoming_relations_async(
                                change.data["dst_address_id"],
                                change.data["src_address_id"],
                            ).result()
                        )
                    )
                    == 1
                ):
                    raise ValueError(
                        f"Updated incoming address relation does not exist: {change}"
                    )
            elif (
                change.action == DbChangeType.UPDATE
                and change.table == "address_outgoing_relations"
            ):
                if not (
                    len(
                        list(
                            tdb.get_address_outgoing_relations_async(
                                change.data["src_address_id"],
                                change.data["dst_address_id"],
                            ).result()
                        )
                    )
                    == 1
                ):
                    raise ValueError(
                        f"Updated outgoing address relation does not exist: {change}"
                    )
            elif (
                change.action == DbChangeType.NEW
                and change.table == "address_incoming_relations"
            ):
                if not (
                    len(
                        list(
                            tdb.get_address_incoming_relations_async(
                                change.data["dst_address_id"],
                                change.data["src_address_id"],
                            ).result()
                        )
                    )
                    == 0
                ):
                    raise ValueError(
                        f"New incoming address relation already exists: {change}"
                    )
            elif (
                change.action == DbChangeType.NEW
                and change.table == "address_outgoing_relations"
            ):
                if not (
                    len(
                        list(
                            tdb.get_address_outgoing_relations_async(
                                change.data["src_address_id"],
                                change.data["dst_address_id"],
                            ).result()
                        )
                    )
                    == 0
                ):
                    raise ValueError(
                        f"New outgoing address relation already exists: {change}"
                    )
            elif (
                change.action == DbChangeType.UPDATE
                and change.table == "cluster_incoming_relations"
            ):
                if not (
                    len(
                        list(
                            tdb.get_cluster_incoming_relations_async(
                                change.data["dst_cluster_id"],
                                change.data["src_cluster_id"],
                            ).result()
                        )
                    )
                    == 1
                ):
                    raise ValueError(
                        f"Updated incoming cluster relation does not exist: {change}"
                    )
            elif (
                change.action == DbChangeType.UPDATE
                and change.table == "cluster_outgoing_relations"
            ):
                if not (
                    len(
                        list(
                            tdb.get_cluster_outgoing_relations_async(
                                change.data["src_cluster_id"],
                                change.data["dst_cluster_id"],
                            ).result()
                        )
                    )
                    == 1
                ):
                    raise ValueError(
                        f"Updated outgoing cluster relation does not exist: {change}"
                    )
            elif (
                change.action == DbChangeType.NEW
                and change.table == "cluster_incoming_relations"
            ):
                if not (
                    len(
                        list(
                            tdb.get_cluster_incoming_relations_async(
                                change.data["dst_cluster_id"],
                                change.data["src_cluster_id"],
                            ).result()
                        )
                    )
                    == 0
                ):
                    raise ValueError(
                        f"New incoming cluster relation already exists: {change}"
                    )
            elif (
                change.action == DbChangeType.NEW
                and change.table == "cluster_outgoing_relations"
            ):
                if not (
                    len(
                        list(
                            tdb.get_cluster_outgoing_relations_async(
                                change.data["src_cluster_id"],
                                change.data["dst_cluster_id"],
                            ).result()
                        )
                    )
                    == 0
                ):
                    raise ValueError(
                        f"New outgoing cluster relation already exists: {change}"
                    )
            elif (
                change.action == DbChangeType.NEW
                and change.table == "cluster_addresses"
            ):
                if not (
                    change.data["cluster_id"] in cluster_seen
                    and change.data["cluster_id"] in cluster_new
                ):
                    raise ValueError(
                        f"Have not seen change creating cluster id: {change}"
                    )
                if not (
                    change.data["address_id"] in addresses_seen
                    and change.data["address_id"] in addresses_new
                ):
                    raise ValueError(f"Have not seen change of address id: {change}")
            elif (
                change.action == DbChangeType.NEW
                and change.table == "address_ids_by_address_prefix"
            ):
                if not (
                    tdb.get_address_id_async(change.data["address"]).result().one()
                    is None
                ):
                    raise ValueError(f"New address already in db: {change}")
                if not (
                    tdb.get_address_async(change.data["address_id"]).result().one()
                    is None
                ):
                    raise ValueError(f"New address_id already in db: {change}")
            elif (
                change.action == DbChangeType.NEW
                and change.table == "summary_statistics"
            ):
                if not (current_summary_stats.no_blocks <= change.data["no_blocks"]):
                    raise ValueError(
                        "Violation: no_blocks db "
                        f"{current_summary_stats.no_blocks} <= "
                        f"{change.data['no_blocks']}"
                    )
                if not (
                    current_summary_stats.no_address_relations
                    <= change.data["no_address_relations"]
                ):
                    raise ValueError(
                        "Violation: no_address_relations db "
                        f"{current_summary_stats.no_address_relations} <= "
                        f"{change.data['no_address_relations']}"
                    )
                if not (
                    current_summary_stats.no_addresses <= change.data["no_addresses"]
                ):
                    raise ValueError(
                        "Violation: no_addresses db "
                        f"{current_summary_stats.no_addresses} <= "
                        f"{change.data['no_addresses']}"
                    )
                if not (
                    current_summary_stats.no_cluster_relations
                    <= change.data["no_cluster_relations"]
                ):
                    raise ValueError(
                        "Violation: no_cluster_relations db "
                        f"{current_summary_stats.no_cluster_relations} <= "
                        f"{change.data['no_cluster_relations']}"
                    )
                if not (
                    current_summary_stats.no_clusters <= change.data["no_clusters"]
                ):
                    raise ValueError(
                        "Violation: no_clusters db "
                        f"{current_summary_stats.no_clusters} <= "
                        f"{change.data['no_clusters']}"
                    )
                if not (
                    current_summary_stats.no_transactions
                    <= change.data["no_transactions"]
                ):
                    raise ValueError(
                        "Violation: no_transactions db "
                        f"{current_summary_stats.no_transactions} <= "
                        f"{change.data['no_transactions']}"
                    )
                # if not (current_summary_stats.timestamp <= change.data["timestamp"]):
                #     raise ValueError(
                #         "Violation: timestamp db "
                #         f"{current_summary_stats.timestamp} <= "
                #         f"{change.data['timestamp']}"
                #     )
            elif change.action == DbChangeType.DELETE:
                raise ValueError(f"Deletes are not allowed: {change}")
            elif (
                change.action == DbChangeType.NEW
                and change.table == "address_transactions"
            ):
                pass
            elif (
                change.action == DbChangeType.NEW
                and change.table == "cluster_transactions"
            ):
                pass
            elif (
                change.action == DbChangeType.NEW
                and change.table == "delta_updater_history"
            ):
                pass
            else:
                raise Exception(f"Have not found validation rule for {change}.")


def apply_changes(
    db: AnalyticsDb, changes: List[DbChange], pedantic: bool, try_atomic_writes: bool
):
    """Apply a list of db-changes to the database. Changes are applied
    atomically and in order.

    Args:
        db (AnalyticsDb): Database instance
        changes (List[DbChange]): List of changes

    Returns:
        None: Nothing

    Raises:
        e: reraises exceptions, a common instance for exceptions is when batches
        are too large. Cassandra has a hard limit for batch sizes
    """

    # Validate changes
    atomic = True
    if pedantic:
        validate_changes(db, changes)

    with LoggerScope.debug(logger, "Summarize updates") as lg:
        if len(changes) == 0:
            lg.debug("Nothing to apply")
            return
        lg.debug(f"{len(changes)} updates to apply. Change Summary:")

        chng_summary = {
            t: group_by(c, lambda y: str(y.action))
            for t, c in group_by(changes, lambda x: x.table).items()
        }
        t_actions = [
            (
                t,
                "; ".join(
                    [
                        f"{a[0].replace('n', '+').replace('d','-')}{len(chngs)}"
                        for a, chngs in actions.items()
                    ]
                ),
            )
            for t, actions in chng_summary.items()
        ]

        short_summary = "; ".join(
            [f"{get_table_abbrev(t)}: {action_str}" for t, action_str in t_actions]
        )
        logger.info(short_summary)

    with LoggerScope.debug(logger, "Applying changes") as _:
        try:
            if try_atomic_writes:
                # try to apply the changes atomic and in-order
                try:
                    db.transformed.apply_changes(changes, atomic=True)
                except InvalidRequest as e:
                    atomic = False
                    msg = getattr(e, "message", repr(e)).lower()
                    if "batch too large" in msg:
                        logger.warning(
                            "Batch to large: Retrying to apply changes "
                            "without atomic write."
                        )
                        db.transformed.apply_changes(changes, atomic=False)
            else:
                atomic = False
                db.transformed.apply_changes(changes, atomic=False)

        except Exception as e:
            atomicity_msg = (
                " Nothing was written (atomic writes)."
                if atomic
                else " Atomicity not guarantied."
            )
            logger.error(f"Failed to apply {len(changes)} changes.{atomicity_msg}")
            raise e


class UpdateStrategyUtxo(UpdateStrategy):
    def __init__(
        self,
        db,
        currency: str,
        pedantic: bool,
        application_strategy: ApplicationStrategy = ApplicationStrategy.TX,
        patch_mode: bool = False,
        forward_fill_rates: bool = False,
    ):
        super().__init__(db, currency, forward_fill_rates=forward_fill_rates)
        crash_file = (
            "/tmp/utxo_deltaupdate_"
            f"{self._db.raw.get_keyspace()}_{self._db.transformed.get_keyspace()}"
            "_crashreport.err"
        )
        stats_value = self._db.transformed.get_summary_statistics()
        """ Make statistics row mutable"""
        self._statistics = (
            MutableNamedTuple(**stats_value._asdict())
            if stats_value is not None
            else DEFAULT_SUMMARY_STATISTICS
        )
        self._pedantic = pedantic
        self._patch_mode = patch_mode
        self.changes = None
        self.application_strategy = application_strategy
        logger.info(f"Updater running in {application_strategy} mode.")
        self.crash_recoverer = CrashRecoverer(crash_file)

    def clear_cache(self):
        pass

    def clear_cache_for_blocks(self, blocks: List[int]):
        pass

    def persist_updater_progress(self):
        if self.changes is not None:
            atomic = ApplicationStrategy.TX == self.application_strategy
            apply_changes(
                self._db, self.changes, self._pedantic, try_atomic_writes=atomic
            )
            self.changes = None
        self._time_last_batch = time.time() - self._batch_start_time

    def prepare_database(self):
        with LoggerScope.debug(logger, "Preparing database"):
            if self._db.transformed.has_delta_updater_v1_tables():
                raise Exception(
                    "Tables of the delta-updater v1 detected. "
                    "please delete new_addresses, dirty_address, "
                    "detla_updater_state and delta-updater_history "
                    "before using delta updater v2."
                )

            HISTORY_TABLE_COLUMNS = [
                "last_synced_block bigint",
                "last_synced_block_timestamp timestamp",
                "highest_address_id int",
                "timestamp timestamp",
                "write_new boolean",
                "write_dirty boolean",
                "runtime_seconds int",
            ]
            HISTORY_TABLE_PK = ["last_synced_block"]

            self._db.transformed.ensure_table_exists(
                TABLE_NAME_DELTA_HISTORY,
                HISTORY_TABLE_COLUMNS,
                HISTORY_TABLE_PK,
                truncate=False,
            )

    def process_batch_impl_hook(self, batch) -> Tuple[Action, Optional[int]]:
        rates = {}
        txs = []
        bts = {}
        """
            Read transaction and exchange rates data
        """
        with LoggerScope.debug(logger, "Checking recovery state.") as lg:
            if self.crash_recoverer.is_in_recovery_mode():
                """
                If we are in recovery mode we start with a block earlier to catch up
                the delta otherwise would start with whats in the db +1
                In case of an error in between blocks this would mean skipping to
                the next block
                """
                mb = max(0, min(batch) - 1)
                lg.warning(
                    "Delta update is in crash recovery mode. Crash hint is "
                    f"{self.crash_recoverer.get_recovery_hint()} in "
                    f"{self.crash_recoverer.get_recovery_hint_filename()} "
                    f" restarting at block {mb}."
                )
                batch = [mb] + batch

        with LoggerScope.debug(logger, "Reading transaction and rates data") as log:
            missing_rates_in_block = False
            for block in batch:
                txs.extend(self._db.raw.get_transactions_in_block(block))
                fiat_values = self._db.transformed.get_exchange_rates_by_block(
                    block
                ).fiat_values
                if fiat_values is None:
                    missing_rates_in_block = True
                    fiat_values = [0, 0]
                rates[block] = fiat_values
                bts[block] = self._db.raw.get_block_timestamp(block)

            if missing_rates_in_block:
                log.warning("Block Range has missing exchange rates. Using Zero.")

        if self.application_strategy == ApplicationStrategy.BATCH:
            if self.crash_recoverer.is_in_recovery_mode():
                raise Exception("Batch mode is not allowed in recovery mode.")
            changes = []
            (
                delta_changes,
                nr_new_addresses,
                nr_new_clusters,
                nr_new_address_relations,
                nr_new_cluster_relations,
            ) = get_transaction_changes(
                self._db,
                txs,
                rates,
                lambda: self.consume_address_id(),
                lambda: self.consume_cluster_id(),
            )

            last_block_processed = batch[-1]
            nr_new_tx = len(txs)
            changes.extend(delta_changes)
            runtime_seconds = int(time.time() - self.batch_start_time)

            changes.extend(
                get_bookkeeping_changes(
                    self._statistics,
                    self._db.transformed.get_summary_statistics(),
                    last_block_processed,
                    nr_new_address_relations,
                    nr_new_addresses,
                    nr_new_cluster_relations,
                    nr_new_clusters,
                    nr_new_tx,
                    self.highest_address_id,
                    runtime_seconds,
                    bts,
                    patch_mode=self._patch_mode,
                )
            )

            # Store changes to be written
            # They are applied at the end of the batch in
            # persist_upater_progress
            self.changes = changes
        elif self.application_strategy == ApplicationStrategy.TX:
            nr_new_address_relations, nr_new_cluster_relations, nr_new_tx = (0, 0, 0)
            last_tx = None
            last_recovery_hint = None
            if self.crash_recoverer.is_in_recovery_mode():
                last_recovery_hint = self.crash_recoverer.get_recovery_hint()
                current_tx = last_recovery_hint["current_tx_id"]
                logger.warning(f"Recovering skipping to tx_id {current_tx}")
                txs = [tx for tx in txs if tx.tx_id >= current_tx]
                self.crash_recoverer.leave_recovery_mode()
            for tx in txs:
                try:
                    if last_recovery_hint is not None:
                        logger.info(
                            f"Resuming processing at tx_id {tx.tx_id} at "
                            f"block {tx.block_id}"
                        )
                    crash_last_succ_tx_id = None
                    crash_last_succ_tx_block_id = None
                    if last_tx is not None:
                        crash_last_succ_tx_id = last_tx.tx_id
                        crash_last_succ_tx_block_id = last_tx.block_id
                    elif last_recovery_hint is not None:
                        """
                        If recovery hint is available and no last tx take over
                        values from there.
                        """
                        crash_last_succ_tx_id = last_recovery_hint[
                            "last_successful_tx_id"
                        ]
                        crash_last_succ_tx_block_id = last_recovery_hint[
                            "last_successful_tx_block_id"
                        ]
                        block_last_error = last_recovery_hint["current_block_id"]
                        assert tx.block_id == block_last_error

                    crash_hint = {
                        "current_block_id": tx.block_id,
                        "current_tx_id": tx.tx_id,
                        "last_successful_tx_id": crash_last_succ_tx_id,
                        "last_successful_tx_block_id": crash_last_succ_tx_block_id,
                    }
                    last_recovery_hint = None
                    with LoggerScope.debug(
                        logger, f"Working on tx_id {tx.tx_id} at block {tx.block_id}"
                    ):
                        with self.crash_recoverer.enter_critical_section(crash_hint):
                            (
                                delta_changes_tx,
                                nr_new_addresses,
                                nr_new_clusters,
                                nr_new_address_relations_tx,
                                nr_new_cluster_relations_tx,
                            ) = get_transaction_changes(
                                self._db,
                                [tx],
                                rates,
                                lambda: self.consume_address_id(),
                                lambda: self.consume_cluster_id(),
                            )
                            last_block_processed = tx.block_id
                            nr_new_tx += 1
                            nr_new_address_relations += nr_new_address_relations_tx
                            nr_new_cluster_relations += nr_new_cluster_relations_tx
                            runtime_seconds = int(time.time() - self.batch_start_time)
                            bookkeepin_changes = get_bookkeeping_changes(
                                self._statistics,
                                self._db.transformed.get_summary_statistics(),
                                last_block_processed,
                                nr_new_address_relations,
                                nr_new_addresses,
                                nr_new_cluster_relations,
                                nr_new_clusters,
                                nr_new_tx,
                                self.highest_address_id,
                                runtime_seconds,
                                bts,
                                patch_mode=self._patch_mode,
                            )
                            apply_changes(
                                self._db,
                                delta_changes_tx + bookkeepin_changes,
                                self._pedantic,
                                try_atomic_writes=True,
                            )
                except Exception as e:
                    assert self.crash_recoverer.is_in_recovery_mode()
                    logger.error(
                        "Entering recovery mode. Recovery hint written "
                        f"at {self.crash_recoverer.get_recovery_hint_filename()}"
                    )
                    logger.error(f"Failed to apply tx {tx.tx_id} ({tx.tx_hash.hex()}).")
                    raise e
                finally:
                    last_tx = tx
        else:
            raise ValueError(
                f"Unknown application strategy {self.application_strategy}"
            )

        return Action.CONTINUE, batch[-1]



import hashlib
import logging
import time
from functools import wraps
from typing import Iterable, List, Optional, Sequence, Union

from cassandra.cluster import (
    EXEC_PROFILE_DEFAULT,
    Cluster,
    ConsistencyLevel,
    ExecutionProfile,
)

# Session,
from cassandra.concurrent import execute_concurrent, execute_concurrent_with_args
from cassandra.policies import DCAwareRoundRobinPolicy, RetryPolicy, TokenAwarePolicy
from cassandra.query import (
    UNSET_VALUE,
    BatchStatement,
    BoundStatement,
    PreparedStatement,
    SimpleStatement,
)

from ..utils import remove_multi_whitespace

DEFAULT_TIMEOUT = 60

# create logger
logger = logging.getLogger(__name__)


# taken from https://github.com/apache/cassandra-dtest/blob/0085d21bc687995478e338302e619e82ad4a4644/dtest.py#L88C5-L88C5 # noqa
class GraphsenseRetryPolicy(RetryPolicy):
    """
    A retry policy that retries 5 times by default, but can be configured to
    retry more times.
    """

    def __init__(self, max_retries=5):
        self.max_retries = max_retries

    def on_read_timeout(self, *args, **kwargs):
        if kwargs["retry_num"] < self.max_retries:
            logger.warning(
                "Retrying read after timeout. Attempt #" + str(kwargs["retry_num"])
            )
            return (self.RETRY, None)
        else:
            return (self.RETHROW, None)

    def on_write_timeout(self, *args, **kwargs):
        logger.warning("write timeout; propagate error to application")
        return (self.RETHROW, None)

    def on_request_error(self, query, consistency, error, retry_num):
        query_string = (
            query.prepared_statement.query_string
            if isinstance(query, BoundStatement)
            else query.query_string
        )
        if query_string.upper().startswith("SELECT ") and retry_num < self.max_retries:
            logger.debug(
                f"Error while executing request; was read; retry on next host: {error}"
            )
            return (self.RETRY_NEXT_HOST, None)
        else:
            logger.debug(
                "Error while executing request; "
                f"propagate error to application: {error}"
            )

        return (self.RETHROW, None)

    # def on_unavailable(self, *args, **kwargs):
    #     if kwargs["retry_num"] < self.max_retries:
    #         logger.debug(
    #             "Retrying request after UE. Attempt #" + str(kwargs["retry_num"])
    #         )
    #         return (self.RETRY, None)
    #     else:
    #         return (self.RETHROW, None)


def normalize_cql_statement(stmt: str) -> str:
    return remove_multi_whitespace(stmt.lower().strip()).rstrip(";")


def none_to_unset(items: Union[dict, tuple, list]):
    """Sets all None value to UNSET to avoid tombstone creation see

    https://stackoverflow.com/questions/34637680/
    how-insert-in-cassandra-without-null-value-in-column

    Args:
        items (Union[dict, tuple, list]): items to insert

    Returns:
        None: -

    Raises:
        Exception: If datatype of items is not supported (list,tuple,dict)
    """
    if type(items) == dict:
        return {k: (UNSET_VALUE if v is None else v) for k, v in items.items()}
    elif type(items) == tuple:
        return tuple([UNSET_VALUE if v is None else v for v in list(items)])
    elif type(items) == list:
        return [(UNSET_VALUE if v is None else v) for v in items]
    else:
        raise Exception(
            f"Can't auto unset for type {type(items)} please assign "
            "cassandra.query.UNSET_VALUE manually."
        )


def get_select_rep(data):
    if type(data) == str or type(data) == int or type(data) == float:
        return str(data)
    else:
        raise Exception(f"No conversion for type {type(data)} to cql select known.")


def get_table_name(table: str, keyspace: Optional[str] = None) -> str:
    return table if keyspace is None else f"{keyspace}.{table}"


def build_select_stmt(
    table: str,
    columns: Sequence[str] = ["*"],
    keyspace: Optional[str] = None,
    where: Optional[dict] = None,
    limit: Optional[int] = None,
    per_partition_limit: Optional[int] = None,
) -> str:
    """Create CQL select statement for specified columns and table name.

    Args:
        table (str): Description
        columns (Sequence[str], optional): Description
        keyspace (Optional[str], optional): Description
        where (Optional[dict], optional): Description
        limit (Optional[int], optional): Description
        per_partition_limit (Optional[int], optional): Description


    Returns:
        str: Description
    """
    cols = ",".join(columns)
    whr = (
        ""
        if where is None
        else " WHERE "
        + " AND ".join([f"{k}={get_select_rep(v)}" for k, v in where.items()])
    )
    lmt = "" if limit is None else f" LIMIT {limit}"
    pplmt = (
        ""
        if per_partition_limit is None
        else f" PER PARTITION LIMIT {per_partition_limit}"
    )
    if per_partition_limit is not None:
        lmt = pplmt
    return f"SELECT {cols} FROM {get_table_name(table, keyspace)}{whr}{lmt};"


def build_insert_stmt(
    columns: Sequence[str], table: str, upsert=True, keyspace: Optional[str] = None
) -> str:
    """Create CQL insert statement for specified columns and table name.

    Args:
        columns (Sequence[str]): column names without type (eg. address)
        table (str): name of the table.

    Returns:
        str: the insert statement
    """

    return (
        f"INSERT INTO {get_table_name(table, keyspace)} ({', '.join(columns)}) "
        f"VALUES ({('?,' * len(columns))[:-1]}) "
        f"{'' if upsert else 'IF NOT EXISTS'}"
        ";"
    )


def build_delete_stmt(
    key_columns: Sequence[str], table: str, keyspace: Optional[str] = None
) -> str:
    """Create CQL insert statement for specified columns and table name.

    Args:
        columns (Sequence[str]): column names without type (eg. address)
        table (str): name of the table.

    Returns:
        str: the insert statement
    """

    return (
        f"DELETE FROM {get_table_name(table, keyspace)} "
        f"WHERE {' AND '.join([ f'{x}=?' for x in key_columns ])}"
        ";"
    )


def build_fail_if_exists_stmt(fail_if_exists: bool) -> str:
    """Summary

    Args:
        fail_if_exists (bool): if statement should fail if item already exists

    Returns:
        str: part of the cql that controls the creation behavior
    """
    return "" if fail_if_exists else "IF NOT EXISTS "


def build_create_stmt(
    columns: Sequence[str],
    pk_columns: Sequence[str],
    table: str,
    fail_if_exists: bool,
    keyspace: Optional[str] = None,
    with_stmt: str = None,
) -> str:
    """Create CQL Create statement

    Args:
        columns (Sequence[str]): Columns with their type (eg. address_prefix text)
        pk_columns (Sequence[str]): Columns used as pk
        table (str): name of the table to create
        fail_if_exists (bool): if an error should be thrown if table already exists

    Returns:
        str: the create statement
    """
    columns_stmt = ", ".join(columns)
    pk_stmt = ",".join(pk_columns)
    w = f" WITH {with_stmt}" if with_stmt is not None else ""

    return (
        f"CREATE TABLE {build_fail_if_exists_stmt(fail_if_exists)}"
        f"{get_table_name(table, keyspace)} "
        f"({columns_stmt}, PRIMARY KEY ({pk_stmt})){w};"
    )


def build_truncate_stmt(table: str, keyspace: Optional[str] = None) -> str:
    if keyspace is None:
        raise Exception("Please provide an explicit keyspace for truncate operations.")
    return f"TRUNCATE {get_table_name(table, keyspace)};"


class StorageError(Exception):
    """Class for Cassandra-related errors"""

    def __init__(self, message: str):
        super().__init__("Cassandra Error: " + message)


class CassandraScope:
    def __init__(self, db_nodes, default_timeout=DEFAULT_TIMEOUT):
        self._db = CassandraDb(db_nodes, default_timeout)

    def __enter__(self):
        self._db.connect()
        return self._db

    def __exit__(self, exc_type, exc_value, tb):
        self._db.close()


class CassandraKeyspaceScope:
    def _set_ks(self, keyspace_name):
        if keyspace_name is not None:
            self._db.set_keyspace(keyspace_name)

    def __init__(self, db, keyspace_name):
        self._db = db
        self._old_keyspace = None
        self._new_keyspace = keyspace_name

    def __enter__(self):
        self._old_keyspace = self._db.get_keyspace()
        self._set_ks(self._new_keyspace)
        return self._db

    def __exit__(self, exc_type, exc_value, tb):
        self._set_ks(self._old_keyspace)


class CassandraDb:
    """Cassandra connector"""

    def needs_session(func):
        @wraps(func)
        def x(*args, **kwargs):
            self = args[0]
            if self.session is None:
                raise StorageError("Session not available. Call connect() first")
            return func(*args, **kwargs)

        return x

    def __init__(self, db_nodes: Iterable, default_timeout=DEFAULT_TIMEOUT) -> None:
        self.db_nodes = db_nodes
        self.cluster = None
        self.session = None
        self.prep_stmts = {}
        self._default_timeout = default_timeout

    def clone(self) -> "CassandraDb":
        return CassandraDb(self.db_nodes, self._default_timeout)

    def __repr__(self):
        return f"{', '.join(self.db_nodes)}"

    def connect(self):
        """Connect to given Cassandra cluster nodes."""
        exec_prof = ExecutionProfile(
            retry_policy=GraphsenseRetryPolicy(),
            consistency_level=ConsistencyLevel.LOCAL_QUORUM,
            serial_consistency_level=ConsistencyLevel.LOCAL_SERIAL,
            request_timeout=self._default_timeout,
            load_balancing_policy=TokenAwarePolicy(DCAwareRoundRobinPolicy()),
        )
        self.cluster = Cluster(
            self.db_nodes,
            execution_profiles={EXEC_PROFILE_DEFAULT: exec_prof},
            connect_timeout=self._default_timeout,
            protocol_version=5,
            compression="lz4",
        )
        try:
            self.session = self.cluster.connect()
        except Exception as e:
            raise StorageError(f"Cannot connect to {self.db_nodes}") from e

    def get_keyspace(self):
        return self.session.keyspace

    def on_keyspace(self, keyspace_name: str):
        return CassandraKeyspaceScope(self, keyspace_name)

    def set_keyspace(self, keyspace_name: str):
        return self.session.set_keyspace(keyspace_name)

    @needs_session
    def set_row_factory(self, factory_function):
        self.session.row_factory = factory_function

    @needs_session
    def has_keyspace(self, keyspace: str) -> bool:
        """Check whether a given keyspace is present in the cluster."""
        try:
            query = "SELECT keyspace_name FROM system_schema.keyspaces"
            result = self.session.execute(query)
            keyspaces = [row.keyspace_name for row in result]
            return keyspace in keyspaces
        except Exception as e:
            raise StorageError(f"Error when executing query: \n{query}") from e

    @needs_session
    def has_table(self, keyspace: str, table: str) -> bool:
        """Check whether a given table is present in the cluster."""
        try:
            query = "SELECT keyspace_name, table_name FROM system_schema.tables"
            result = self.session.execute(query)
            tables = [row.table_name for row in result if row.keyspace_name == keyspace]
            return table in tables
        except Exception as e:
            raise StorageError(f"Error when executing query: \n{query}") from e

    @needs_session
    def execute_safe(
        self, cql_query_str: str, params: dict, fetch_size=None
    ) -> Iterable:
        # flat_stmt = cql_query_str.replace("\n", " ")
        # logger.debug(f"{flat_stmt} in keyspace {self.session.keyspace}")
        stmt = SimpleStatement(cql_query_str, fetch_size=fetch_size)
        return self.session.execute(stmt, params)

    @needs_session
    def execute(self, cql_query_str: str, fetch_size=None) -> Iterable:
        # flat_stmt = cql_query_str.replace("\n", " ")
        # logger.debug(f"{flat_stmt} in keyspace {self.session.keyspace}")

        stmt = SimpleStatement(cql_query_str, fetch_size=None)
        if fetch_size is not None:
            stmt.fetch_size = fetch_size

        return self.session.execute(stmt)

    @needs_session
    def execute_statement(self, stmt: BoundStatement, fetch_size=None) -> Iterable:
        return self.session.execute(stmt)

    @needs_session
    def execute_async(self, cql_query_str: str, fetch_size=None):
        # flat_stmt = cql_query_str.replace("\n", " ")
        # logger.debug(f"{flat_stmt} in keyspace {self.session.keyspace}")
        stmt = SimpleStatement(cql_query_str, fetch_size=None)
        if fetch_size is not None:
            stmt.fetch_size = fetch_size
        return self.session.execute_async(stmt)

    @needs_session
    def execute_async_safe(self, cql_query_str: str, params: dict, fetch_size=None):
        # flat_stmt = cql_query_str.replace("\n", " ")
        # logger.debug(f"{flat_stmt} in keyspace {self.session.keyspace}")
        stmt = SimpleStatement(cql_query_str, fetch_size=fetch_size)
        return self.session.execute_async(stmt, params)

    def execute_statements_atomic(self, statements: List[BoundStatement]):
        batch = BatchStatement()
        for stmt in statements:
            batch.add(stmt)

        self.session.execute(batch)

    @needs_session
    def execute_statements(self, statements: List[BoundStatement]):
        return execute_concurrent(
            self.session,
            [(stmt, None) for stmt in statements],
            raise_on_first_error=True,
        )

    @needs_session
    def execute_statements_async(
        self, statements: List[BoundStatement], concurrency=100
    ):
        return execute_concurrent(
            self.session,
            [(stmt, None) for stmt in statements],
            raise_on_first_error=True,
            concurrency=concurrency,
            results_generator=True,
        )

    @needs_session
    def execute_statement_async(self, stmt, params):
        return self.session.execute_async(stmt, params)

    @needs_session
    def execute_batch_async(self, stmt, params):
        if len(params) > 10000:
            logger.warning(
                "CAUTION: Running many (10k+) parallel queries against db "
                "without concurrency control. "
                "Might lead to timeouts. Consider using execute_statements_async."
            )
        prp = self.get_prepared_statement(stmt)
        futures = [
            (identifier, self.execute_statement_async(prp, param_list))
            for identifier, param_list in params
        ]
        return futures

    @needs_session
    def execute_batch(self, stmt, params):
        return [
            (i, future.result())
            for (i, future) in self.execute_batch_async(stmt, params)
        ]

    def await_batch(self, futures):
        return [(i, future.result()) for (i, future) in futures]

    @needs_session
    def setup_keyspace_using_schema(self, schema: str) -> None:
        """Setup keyspace and tables."""

        statements = schema.split(";")
        for stmt in statements:
            stmt = stmt.strip()
            if len(stmt) > 0:
                self.session.execute(stmt + ";")

    @needs_session
    def get_columns_for_table(self, keyspace: str, table: str):
        cql_str = (
            f"SELECT * FROM system_schema.columns "
            f"WHERE keyspace_name = '{keyspace}' "
            f"AND table_name = '{table}';"
        )
        return self.session.execute(cql_str)

    @needs_session
    def get_prepared_insert_statement(
        self, keyspace: str, table: str, upsert=True, cl=None
    ) -> PreparedStatement:
        """Build prepared CQL statement for specified table."""

        result_set = self.get_columns_for_table(keyspace, table)
        columns = [elem.column_name for elem in result_set._current_rows]
        ps = self.get_prepared_statement(
            build_insert_stmt(columns, table, upsert=upsert, keyspace=keyspace)
        )
        if cl is not None:
            ps.consistency_level = cl
        return ps

    @needs_session
    def get_prepared_statement(self, stmt):
        hash_object = hashlib.sha256(stmt.encode("utf-8"))
        hex_dig = hash_object.hexdigest()
        if hex_dig not in self.prep_stmts:
            self.prep_stmts[hex_dig] = self.session.prepare(stmt)
        return self.prep_stmts[hex_dig]

    @needs_session
    def ingest(
        self,
        table: str,
        keyspace: str,
        items: Iterable,
        upsert=True,
        cl=None,
        concurrency: int = 100,
        auto_none_to_unset=False,
    ):
        stmt = self.get_prepared_insert_statement(keyspace, table, upsert=upsert, cl=cl)

        if auto_none_to_unset:
            items = [none_to_unset(row) for row in items]

        self._exe_with_retries(stmt, items, concurrency=concurrency)

    @needs_session
    def _exe_with_retries(
        self,
        prepared_stmt: PreparedStatement,
        parameters,
        concurrency: int = 100,
        retries=4,
    ) -> None:
        """Concurrent ingest into Apache Cassandra."""
        nr_statements = len(parameters)
        retries = (retries - 1) * nr_statements

        while True:
            try:
                results = execute_concurrent_with_args(
                    session=self.session,
                    statement=prepared_stmt,
                    parameters=parameters,
                    concurrency=concurrency,
                )

                for i, (success, _) in enumerate(results):
                    if not success:
                        while True:
                            try:
                                logger.warning(
                                    "Retrying failed statement:",
                                    prepared_stmt,
                                    parameters[i],
                                )
                                self.session.execute(prepared_stmt, parameters[i])
                            except Exception as exception:
                                logger.warning(
                                    f"Retry after exception: {str(exception)} "
                                    f"retrying another {retries} times/statements"
                                )
                                retries -= 1
                                if retries < 0:
                                    logger.error("Giving up retries for ingest.")
                                    raise exception
                                else:
                                    continue
                            break
                break

            except Exception as exception:
                logger.warning(
                    f"Retry after exception: {str(exception)} "
                    f"retrying another {retries} times/statements"
                )
                # TODO: Refactor, sleeps are probably only masking another problem
                time.sleep(1)
                retries -= nr_statements
                if retries < 0:
                    logger.error("Giving up retries for ingest.")
                    raise exception
                else:
                    continue

    @needs_session
    def setup_keyspace_using_schema_file(self, schema_file: str) -> None:
        with open(schema_file, "r", encoding="utf-8") as file_handle:
            self.setup_keyspace_using_schema(file_handle.read())

    @needs_session
    def close(self):
        """Closes the cassandra cluster connection."""
        self.cluster.shutdown()



from typing import List

from ..utils.slack import send_message_to_slack
from .abstract import EventNotifier, FlowEvent, WatchConfig


class SlackEventNotifier(EventNotifier):
    def __init__(self, slack_webhook_urls: List[str]):
        self.slack_webhook_urls = slack_webhook_urls
        self.msgs = []

    def add_notification(self, flow: FlowEvent, receiver_config: WatchConfig, raw_tx):
        self.msgs.append(f"{receiver_config.email}{raw_tx}")

    def send_notifications(self):
        for msg in self.msgs:
            for hook in self.slack_webhook_urls:
                send_message_to_slack(f"{msg}", hook)
        self.msgs = []



import os

from ..config import (
    ConfigError,
    avg_blocktimes_by_currencies,
    config,
    currency_to_schema_type,
)
from ..db import DbFactory
from ..ingest.utxo import CassandraOutputResolver
from .account import EthereumEtlFlowProvider
from .flatfile import JsonWatcherState, JsonWatchpointProvider
from .logging import LoggingEventNotifier
from .slack import SlackEventNotifier
from .utxo import BitcoinEtlFlowProvider
from .watcher import FlowWatcher


class FlowWatcherFactory:
    def file_based_from_config(
        self, env, currency, state_file, watchpoints_file
    ) -> FlowWatcher:
        if not os.path.isfile(watchpoints_file):
            raise ValueError(f"Watchpoints file ({watchpoints_file}) not found.")
        schema = currency_to_schema_type[currency]

        notifiers = [LoggingEventNotifier()]
        slack_topic = config.get_slack_hooks_by_topic("payment_flow_notifications")
        if slack_topic:
            notifiers.append(SlackEventNotifier(slack_topic.hooks))

        ks_config = config.get_keyspace_config(env, currency)
        if (
            ks_config.ingest_config is None
            or ks_config.ingest_config.get_first_node_reference() is None
        ):
            raise ConfigError(
                "There is no node_reference specified in the config "
                f"({env}.{currency}.ingest_config.node_reference is missing)"
            )
        node_ref = ks_config.ingest_config.get_first_node_reference()

        return FlowWatcher(
            state=JsonWatcherState(state_file),
            watchpoints=JsonWatchpointProvider(watchpoints_file),
            flow_provider=(
                EthereumEtlFlowProvider(node_ref)
                if schema == "account"
                else BitcoinEtlFlowProvider(
                    currency,
                    node_ref,
                    CassandraOutputResolver(DbFactory().from_config(env, currency)),
                )
            ),
            notifiers=notifiers,
            new_block_backoff_sec=avg_blocktimes_by_currencies.get(currency, 600),
        )



# flake8: noqa: F501
import importlib
import random
from functools import reduce

import pytest

from graphsenselib.utils.address import InvalidAddress, address_to_bytes, address_to_str

from . import resources

testset_correct = [
    (
        "trx",
        "TGCRkw1Vq759FBCrwxkZGgqZbRX1WkBHSu",
        None,
        "4450cf8c8b6a8229b7f628e36b3a658e84441b6f",
    ),
    (
        "TRX",
        "TWsm8HtU2A5eEzoT8ev8yaoFjHsXLLrckb",
        None,
        "e552f6487585c2b58bc2c9bb4492bc1f17132cd0",
    ),
    (
        "trx",
        "THKJYuUmMKKARNf7s2VT51g5uPY6KEqnat",
        None,
        "5095d4f4d26ebc672ca12fc0e3a48d6ce3b169d2",
    ),
    (
        "eth",
        "4450cf8c8b6a8229b7f628e36b3a658e84441b6f",
        None,
        "4450cf8c8b6a8229b7f628e36b3a658e84441b6f",
    ),
    (
        "eth",
        "e552f6487585c2b58bc2c9bb4492bc1f17132cd0",
        None,
        "e552f6487585c2b58bc2c9bb4492bc1f17132cd0",
    ),
    (
        "eth",
        "5095d4f4d26ebc672ca12fc0e3a48d6ce3b169d2",
        None,
        "5095d4f4d26ebc672ca12fc0e3a48d6ce3b169d2",
    ),
    (
        "eth",
        "0x4450cf8c8b6a8229b7f628e36b3a658e84441b6f",
        "4450cf8c8b6a8229b7f628e36b3a658e84441b6f",
        "4450cf8c8b6a8229b7f628e36b3a658e84441b6f",
    ),
    (
        "eth",
        "0xe552f6487585c2b58bc2c9bb4492bc1f17132cd0",
        "e552f6487585c2b58bc2c9bb4492bc1f17132cd0",
        "e552f6487585c2b58bc2c9bb4492bc1f17132cd0",
    ),
    (
        "eth",
        "0x5095d4f4d26ebc672ca12fc0e3a48d6ce3b169d2",
        "5095d4f4d26ebc672ca12fc0e3a48d6ce3b169d2",
        "5095d4f4d26ebc672ca12fc0e3a48d6ce3b169d2",
    ),
    (
        "eth",
        "0xFF4c1897369C0DFa37e3442BB08c810A33214349",
        "FF4c1897369C0DFa37e3442BB08c810A33214349".lower(),
        "ff4c1897369c0dfa37e3442bb08c810a33214349",
    ),
    (
        "bch",
        "bitcoincash:pvhfadjrmc4adxl8payajshf34wauh4jlrfnpuyetfnzh5llrnh9w3q8y9qcp",
        "B5Fj9EniQGXUKCCx1dfm7XXARMf38wBgtjPTQv9uaR24eTrj3T",
        "2c53eb24ebaa6107dc4cc3380659ed1df7ca6559c32372e8d2b5db63627589908499bcab0db0",
    ),
    (
        "bch",
        "bitcoincash:pvqljqfuvj0dqf4l6s9mradymsvmucwe8w7cm84yyt8p2dtwj93xcuqwaelt6",
        "AjbGjygbRL7MwFc3PJ9KWVqMdUGeHsFxXDaiGzyTqQnBKP4iWo",
        "2ab8d0af9a236541d5dcf9035d225379dc5595c4264733f87cd8aa43ae5bc58b8b4d712a7af0",
    ),
    (
        "bch",
        "bitcoincash:pv4dcn3jppealt8y9t6egv4aj8snhtt4qqz006u27dezdwj65xkxuqaqmccfz",
        "B3bhUDrC913vH9rRaxnhToG4fi1hp8B55Sg1RHGE4AdCRoti1o",
        "2c38e970dc8c2410f6449c998b8ba96ef4049ea069c082c515aa0165140e10a94c66fd2a06f0",
    ),
    (
        "bch",
        "B5Fj9EniQGXUKCCx1dfm7XXARMf38wBgtjPTQv9uaR24eTrj3T",
        None,
        "2c53eb24ebaa6107dc4cc3380659ed1df7ca6559c32372e8d2b5db63627589908499bcab0db0",
    ),
    (
        "bch",
        "AjbGjygbRL7MwFc3PJ9KWVqMdUGeHsFxXDaiGzyTqQnBKP4iWo",
        None,
        "2ab8d0af9a236541d5dcf9035d225379dc5595c4264733f87cd8aa43ae5bc58b8b4d712a7af0",
    ),
    (
        "bch",
        "B3bhUDrC913vH9rRaxnhToG4fi1hp8B55Sg1RHGE4AdCRoti1o",
        None,
        "2c38e970dc8c2410f6449c998b8ba96ef4049ea069c082c515aa0165140e10a94c66fd2a06f0",
    ),
    (
        "btc",
        "1ND5TQ2AnQatxqmHrdMT8utb3aWgLKsc9w",
        None,
        "0563456d808ab988b4e31b51ca555b235d230e27a8513ce42770",
    ),
    (
        "btc",
        "18mBbKWTGAbdbP2h24Nbx5AtjWLxScJP94",
        None,
        "048b4b8d379b40a8e58d70a90845a3e052b4ade5386a44972440",
    ),
    (
        "btc",
        "19zUcPa1ENfnSy8RMJ3AbF4TXMaZrskKKb",
        None,
        "049e9c9178813969ee6b92195520ca8cf11b7d58a1cb3b134e30",
    ),
    (
        "btc",
        "1C3gCLqhxPb8asjtuLBe6dF5FzXRo5nANe",
        None,
        "04c0e8314c69e178c88b3af4d542e61a53c53fa7d9bc5b8a5a60",
    ),
    (
        "btc",
        "bc1ql8dxppm0ge7g3nx4x9l4lssazf9j2wxzxq8xcx",
        None,
        "85988180838708271025a7c94941d61c681681145e0ca1932cf1c31c12076470",
    ),
    (
        "btc",
        "36TcL12bmahTRrd3dDLRRLQBHYp64xUdng",
        None,
        "0c66e45010a3b62a5b67294394d51965460b460c06138725ba80",
    ),
    (
        "btc",
        "3GCsjdaiebDCBFAoiVv4odtP1dmruRBJCy",
        None,
        "0d0333ae58aa9a334c2cf2afa9dd84be5d17065b72d592d23390",
    ),
    (
        "ltc",
        "LiM295pXpydyt4ahLxEYJeevpLtm2e6LSv",
        None,
        "52a542245c1fc39979d048a95383a04a69b6c14d2d0a61946b60",
    ),
    (
        "ltc",
        "ltc1qq4zfq0p9s5e69y4vp0kx5gvjvpku8zz7c6q9ga",
        None,
        "80c66204158328140219155a6c61563424171d524d4cd0977480c37d96c1189780",
    ),
    (
        "ltc",
        "ltc1qmrf5gsptkp2xsunvxc57txjhr53w4d5003smpe",
        None,
        "80c66205c10a54944231708b1d17543476557cc1d36045523d6395410491702680",
    ),
    (
        "btc",
        "17VZNX1SN5NtKa8UQFxwQbFeFc3iqRYhem",
        None,
        "04776159f05a5855b44e221c60fe376233e63e40eac598299ad0",
    ),
    (
        "btc",
        "3EktnHQD7RiAE6uzMj2ZifT9YgRrkSgzQX",
        None,
        "0ceb34b9160d1d9a8a386d7a56b0a1aa76c9828672b1aa3a61f0",
    ),
    (
        "btc",
        "5Hwgr3u458GLafKBgxtssHSPqJnYoGrSzgQsPwLFhLNYskDPyyA",
        None,
        "151de8c83d441484148a74cba38d33cd1697c52ba0bd0c9aea86335f750fa545a0cec357e79280",
    ),  # Private key (WIF, uncompressed pubkey)
    (
        "btc",
        "5TfQjD9DLFeUFmDiDrzsdtSGQss93o4pvsmQcgmjfcQVLsEgAoM",
        None,
        "15b9d8acd24d50f99c3ed36a372eb3974690633cc90ef130db3b58928b6b9e461d5333a82af540",
    ),  # Private key (WIF, uncompressed pubkey, Electrum-defined and now deprecated)
    (
        "btc",
        "L1aW4aubDFB7yfras2S1mN3bqg9nwySY8nkoLmJebSLD5BWv3ENZ",
        None,
        "50189e122d6334f2c7e67ca2cc2681b560e3c6826edf96a022eb2f52d4a68da50d14b7b60ce5a1",
    ),  # Private key (WIF, compressed pubkey)
    (
        "btc",
        "xpub661MyMwAqRbcEYS8w7XLSVeEsBXy79zSzH1J8vCdxAZningWLdN3",
        None,
        "e30d63186055e55dcac598e43a0688dc77d469d98eccb7f91c9e9ae91052236325e0a86eaaea1e525583",
    ),  # BIP32 pubkey
    (
        "btc",
        "xprv9s21ZrQH143K24Mfq5zL5MhWK9hUhhGbd45hLXo2Pq2oqzMMo63o",
        None,
        "e30cb62730818726110440d3084567c45e94155a5e4c9a5ca6942394416951fbc25f10afc7a555bc60ef",
    ),  # BIP32 private key
    (
        "btc",
        "cNJFgo1driFnPcBdBX8BrJrpxchBWXwXCvNH5SoSkdcF6JXXwHMm",
        None,
        "91648fa2f065caa3ee5e42e52df20bc92cb0e24a4b79fddf33659115abdab2590f1927dfdd156d",
    ),  # Testnet Private key (WIF, compressed pubkey)
    (
        "btc",
        "tpubD6NzVbkrYhZ4WLczPJWReQycCJdd6YVWXubbVUFnJ5KgU5MDQrD9",
        None,
        "d30d633465ba763b3282984479493a5d2799998e643129651a075e7f58e375c3ee4854e870554d632349",
    ),  # Testnet BIP32 pubkey
    (
        "btc",
        "bc1qw508d6qejxtdg4y5r3zarvary0c5xw7kv8f3t4",
        None,
        "8598813d540839b05a4c730e2561551120de10d7841506551cf7d73482923160",
    ),  # Bech32 pubkey hash or script hash
    (
        "btc",
        "bc1qw508d6qejxtdg4y5r3zarvary0c5xw7kxpjzsx",
        None,
        "8598813d540839b05a4c730e2561551120de10d7841506551cf7d71c24c34470",
    ),  # Bech32 pubkey hash or script hash
]


def test_address_converter():
    compression_btclike = []
    for n, a, ra, ob in testset_correct:
        b = address_to_bytes(n, a)
        c = address_to_str(n, b)
        if ra is None:
            assert a == c
        else:
            assert ra == c
        assert b.hex() == ob

        if n != "trx" and n != "eth":
            compression_btclike.append(len(b) / len(a.encode("utf-8")))

    avg_compression_btclike = sum(compression_btclike) / len(compression_btclike)

    assert avg_compression_btclike < 0.75


def test_address_converter_inject_nonencodable_char():
    rpmts = [
        ("0", "!"),
        ("1", "!"),
        ("V", "!"),
        ("a", "!"),
        ("5", "!"),
        ("j", "!"),
        ("q", "!"),
    ]

    for n, a, ra, ob in testset_correct:
        ac = reduce(lambda a, kv: a.replace(*kv), rpmts, a)
        with pytest.raises(InvalidAddress):
            b = address_to_bytes(n, ac)


@pytest.mark.slow
def test_address_converter_testset():
    tests = 0
    pct = 0.8
    networks = ["bch", "btc", "ltc", "zec"]
    addresses_per_network = 100000
    pct_error = len(networks) * addresses_per_network * (0.01)
    e_tests = len(networks) * addresses_per_network * (1 - pct) - pct_error

    for network in networks:
        with importlib.resources.path(
            resources,
            f"{network}_addresses.txt",
        ) as path:
            with open(path) as f:
                for a in f.readlines():
                    if random.uniform(0, 1) > pct:
                        a = a.strip()
                        b = address_to_bytes(network, a)
                        c = address_to_str(network, b)
                        assert a == c
                        tests += 1

    assert tests > e_tests
